/*
 * chan_oh323.c
 *
 * Generic OpenH323 Channel Driver for ASTERISK PBX.
 * 
 * Copyright (C) 2002,2003 Inaccess Networks.
 * Michael Manousos <manousos@inaccessnetworks.com>
 *
 * This file is part of "H323 support for ASTERISK"
 *
 * "H323 support for ASTERISK" is free software;
 * you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any later version. 
 *
 * "H323 support for ASTERISK" is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 * In-Band DTMF detection is based on code sent by Klaus-Peter Junghanns
 * in ASTERISK mailing list.
 *
 * $Id: chan_oh323.c,v 1.43 2003/05/09 16:19:06 manousos Exp $
 *
 */

/********************************************************************************/
/* System and application's include files ***************************************/
#include <stdio.h>
#include <pthread.h>
#include <string.h>
#include <asterisk/channel.h>
#include <asterisk/channel_pvt.h>
#include <asterisk/config.h>
#include <asterisk/logger.h>
#include <asterisk/module.h>
#include <asterisk/pbx.h>
#include <asterisk/cli.h>
#include <asterisk/options.h>
#include <asterisk/dsp.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <math.h>
#include <netinet/ip.h>

#include "wrapper.hxx"
#include "oh323.h"

/********************************************************************************/
/* String variables required by ASTERISK ****************************************/
static char *type = "OH323";
static char *desc = "OpenH323 Channel Driver";
static char *tdesc = "OpenH323 Channel Driver";
static char *config = "oh323.conf";

/********************************************************************************/
/* Channel driver's internal stuff **********************************************/

/* Some function prototypes of the channel driver */
static void	oh323_destroy(int);
static char	*build_fifo_name(lchan_dir_t, int);
static int	copy_call_details(call_details_t *, call_details_t *);
static void	clear_call_details(call_details_t *);
static int	oh323_codec2format(int);
/*static char *oh323_codec2str(int);*/
static int	oh323_format2codec(int);
static char	*oh323_state2str(int);
static char	*oh323_capability2str(int, char**);
static char *oh323_format2str(int);
static int context_from_alias(char *, char **);
static int context_from_prefix(char *, char **);

/* Statistics of channel's usage */
typedef struct {
    int incall;
    int outcall;
	int setup_sent;
	int setup_recv;
	int call_error;
	int block_incall;
	int block_outcall;
} OH323STATS;

/* Enable/Disable verbose debug info */
static int	oh323_verbose_debug = 0;

/** The capabilities (codecs) supported by H.323 channel driver */
static int oh323_capability = OH323_CAPABILITY_FULLBANDWIDTH;

/** Static buffer for fifo name creation (out-dated) */
static char name[MAX_FIFO_NAME];

/** Mutex for the "name" buffer (out-dated) */
static pthread_mutex_t name_lock = PTHREAD_MUTEX_INITIALIZER;

/** Mutex for the usage counter */
static pthread_mutex_t usecnt_lock = PTHREAD_MUTEX_INITIALIZER;
static int usecnt = 0;

/** Mutex for the OpenH323 calls table */
static pthread_mutex_t oh323_tab_lock = PTHREAD_MUTEX_INITIALIZER;

/** Mutex for the OpenH323 statistics structure */
static pthread_mutex_t oh323_stats_lock = PTHREAD_MUTEX_INITIALIZER;
static OH323STATS oh323_stats;

/*
 * Structures for storing configuration 
 * options of H.323 channel driver.
 */
struct oh323_reginfo {
	char context[AST_MAX_EXTENSION];
	char **alias;
	int  alias_num;
	char **prefix;
	int  prefix_num;
	struct oh323_reginfo *next;
};

struct oh323_codecinfo {
	int codec;
	int format;
	int frames;
	struct oh323_codecinfo *next;
};

struct oh323_options {
	char listenAddress[128];
	int  listenPort;
	int  connectPort;
	int  fastStart;
	int  h245Tunnelling;
	int  h245inSetup;
	int  inBandDTMF;
	int  silenceSuppression;
	int  jitterMin;
	int  jitterMax;
	int  ipTos;
	int  outboundMax;
	int  inboundMax;
	int  totalNum;
	int  libTraceLevel;
	char libTraceFile[256];
	int  amaFlags;
	char accountCode[20];
	gkmode_t gatekeeperMode;
	char gatekeeperName[128];
	char gatekeeperPass[128];
	uimode_t userInputMode;
	char context[AST_MAX_EXTENSION];
	struct oh323_reginfo *regInfo;
	struct oh323_codecinfo *codecInfo;
};
static struct oh323_options config_options;

/** Structure for transfering data occurred inside 
 * an exception of OpenH323 Library. */
struct exception_oh323 {

	/** Type of exception data. */
	int type;

	/** Buffer with exception data */
	char data[MAX_EXCEPTION_DATA];
};

/**
 * Private structure of a OpenH323 channel.
 */
struct chan_oh323_pvt {

	/** Channel's file descriptors. The two arrays hold 
	 * 2 socketpairs. The first member of each array (index 0)
	 * is used inside the channel driver, whereas the 
	 * second member (index 1) is used inside the OpenH323.
	 * player_fd[0] is used to read data from the network
	 * (OpenH323 library) and record_fd[0] is used to write
	 * data to the network. player_fd[0] is select()'ed 
	 * inside ASTERISK. */
	int	player_fd[2];
	int recorder_fd[2];

	/** Player fifo name (data flow: OH323 --> ASTERISK) */
	char playername[MAX_FIFO_NAME];

	/** Recorder fifo name (data flow: ASTERISK --> OH323) */
	char recordername[MAX_FIFO_NAME];

	/** Options to be used during call setup */
	call_options_t call_opt;

	/** Call details */
	call_details_t cd;

	/** Direction established */
	lchan_dir_t direction;

	struct ast_dsp *vad;

	/** The capability (codec) used by this channel */
	int capability;
	
	/** Buffer sizes for RX/TX */
	int rx_buf_size, tx_buf_size;

	/** Owner if we have one */
	struct ast_channel *owner;

	/** Asterisk frame */
	struct ast_frame fr;

	/** Smoother struct for queuing frames for transmission */
	struct ast_smoother *tx_smooth;

	/** Static buffer for reading frames */
	char buf[OH323_MAX_BUF];

	/** State indication (OH323_STATE_*) */
	int i_state;

	/** Flag indicating whether the call was initiated from
	 * the remote endpoint or from ASTERISK.
	 */
	int from_remote;

	/** Structure containing exception data */
	struct exception_oh323 except_struct;

	/** Already gone */
	int	alreadygone;
};

/**
 * Array of OpenH323 calls (either established or inactive).
 */
static struct chan_oh323_pvt **oh323_tab;

/*
 * CLI extensions 
 */
static int oh323_show_conf(int fd, int argc, char *argv[])
{
	char	*tmp;
	char	tmp_gk[200];

	if (argc != 3)
		return RESULT_SHOWUSAGE;
	ast_pthread_mutex_lock(&oh323_stats_lock);
	ast_cli(fd, "\n");
	ast_cli(fd, "Configuration of OpenH323 channel driver\n");
	ast_cli(fd, "----------------------------------------\n");
	ast_cli(fd, "Version: %d.%d.%d\n", OH323_VERSION_MAJOR, OH323_VERSION_MINOR, 
					OH323_VERSION_BUILD);
	ast_cli(fd, "Listening on address: %s:%d\n", config_options.listenAddress, 
					config_options.listenPort);
	ast_cli(fd, "Gatekeeper used: %s\n", h323_get_gk(tmp_gk));
	ast_cli(fd, "FastStart/H245Tunnelling/H245inSetup: %s/%s/%s\n", 
					config_options.fastStart!=0?"ON":"OFF",
					config_options.h245Tunnelling!=0?"ON":"OFF",
					config_options.h245inSetup!=0?"ON":"OFF");
	ast_cli(fd, "Supported format(s): %s\n", oh323_capability2str(oh323_capability, &tmp));
	ast_cli(fd, "Jitter buffer limits (min/max): %d-%d\n", config_options.jitterMin,
					config_options.jitterMax);
	ast_cli(fd, "IP Type-of-Service value: %d\n", config_options.ipTos);
	ast_cli(fd, "User input mode: %d\n", config_options.userInputMode);
	ast_cli(fd, "Max number of inbound H.323 calls: %d\n", config_options.inboundMax);
	ast_cli(fd, "Max number of outbound H.323 calls: %d\n", config_options.outboundMax);
	ast_cli(fd, "\n");
	if (tmp != NULL)
		free(tmp);
	ast_pthread_mutex_unlock(&oh323_stats_lock);
	return RESULT_SUCCESS;
}
static int oh323_show_stats(int fd, int argc, char *argv[])
{
	if (argc != 3)
		return RESULT_SHOWUSAGE;
	ast_pthread_mutex_lock(&oh323_stats_lock);
	ast_cli(fd, "\n");
	ast_cli(fd, "Statistics of OpenH323 channel driver\n");
	ast_cli(fd, "-------------------------------------\n");
	ast_cli(fd, "Inbound H.323 calls: %-6d\n", oh323_stats.incall);
	ast_cli(fd, "Outbound H.323 calls: %-6d\n", oh323_stats.outcall);
	ast_cli(fd, "Dropped inbound H.323 calls: %-6d\n", oh323_stats.block_incall);
	ast_cli(fd, "Blocked outbound H.323 calls: %-6d\n", oh323_stats.block_outcall);
	ast_cli(fd, "Total inbound H.323 calls detected: %-6d\n", oh323_stats.setup_recv);
	ast_cli(fd, "Total outbound H.323 calls attempted: %-6d\n", oh323_stats.setup_sent);
	ast_cli(fd, "\n");
	ast_pthread_mutex_unlock(&oh323_stats_lock);
	return RESULT_SUCCESS;
}
static int oh323_show_info(int fd, int argc, char *argv[])
{
	int		i, count;

	if (argc != 3)
		return RESULT_SHOWUSAGE;
	count = 0;
	ast_pthread_mutex_lock(&oh323_tab_lock);
	ast_cli(fd, "\n");
	ast_cli(fd, "Information about active OpenH323 channel(s)\n");
	ast_cli(fd, "--------------------------------------------\n");
	for (i=0; i<config_options.totalNum; i++)
		if (oh323_tab[i] != NULL) {
			ast_cli(fd, "%-2d: ", count);
			if (oh323_tab[i]->cd.call_token != NULL)
				ast_cli(fd, "token=%s, ", oh323_tab[i]->cd.call_token);
			ast_cli(fd, "state=%s, ", oh323_state2str(oh323_tab[i]->i_state));
			ast_cli(fd, "from_remote=%d, ", oh323_tab[i]->from_remote);
			ast_cli(fd, "RX/TX=%d/%d, ", oh323_tab[i]->rx_buf_size, oh323_tab[i]->tx_buf_size);
			ast_cli(fd, "format=%s, ", oh323_format2str(oh323_tab[i]->capability));
			++count;
			ast_cli(fd, "\n");
		}
	if (count == 0)
		ast_cli(fd, "  <No active H.323 connections>\n");
	ast_cli(fd, "\n");
	ast_pthread_mutex_unlock(&oh323_tab_lock);
	return RESULT_SUCCESS;
}
static int oh323_debug_toggle(int fd, int argc, char *argv[])
{
	if (argc != 3)
		return RESULT_SHOWUSAGE;
	ast_pthread_mutex_lock(&oh323_tab_lock);
	if (oh323_verbose_debug != 0) {
		oh323_verbose_debug = 0;
		ast_cli(fd, "Verbose debug info for OpenH323 channel turned off.\n");
	} else {
		oh323_verbose_debug = 1;
		ast_cli(fd, "Verbose debug info for OpenH323 channel turned on.\n");
	}
	ast_pthread_mutex_unlock(&oh323_tab_lock);
	return RESULT_SUCCESS;
}

static char oh323_show_conf_usage[] =
"Usage: oh323 show conf\n"
"       Show config info of OpenH323 channel driver\n";
static char oh323_show_stats_usage[] =
"Usage: oh323 show stats\n"
"       Show statistics of OpenH323 channel driver\n";
static char oh323_show_info_usage[] =
"Usage: oh323 show info\n"
"       Show verbose information about active OpenH323 channel(s)\n";
static char oh323_debug_toggle_usage[] =
"Usage: oh323 debug toggle\n"
"       Toggle on/off verbose debug info for OpenH323 channel driver\n";
static struct ast_cli_entry	cli_show_conf = 
	{ { "oh323", "show", "conf", NULL }, oh323_show_conf, 
			"Show config info of OpenH323 channel driver", oh323_show_conf_usage };
static struct ast_cli_entry	cli_show_stats = 
	{ { "oh323", "show", "stats", NULL }, oh323_show_stats, 
			"Show statistics of OpenH323 channel driver", oh323_show_stats_usage };
static struct ast_cli_entry	cli_show_info = 
	{ { "oh323", "show", "info", NULL }, oh323_show_info, 
			"Show info about active OpenH323 channel(s)", oh323_show_info_usage };
static struct ast_cli_entry	cli_debug_toggle = 
	{ { "oh323", "debug", "toggle", NULL }, oh323_debug_toggle, 
			"Toggle on/off debug info for OpenH323 channel driver", oh323_debug_toggle_usage };


/**
 * Handle exceptions of the OpenH323 channel.
 * Currently only received DTMF tones are handled.
 */
static struct ast_frame *oh323_exception(struct ast_channel *c)
{
	struct chan_oh323_pvt *p = c->pvt->pvt;

	if (option_debug)
		ast_log(LOG_DEBUG, "In oh323_exception (%s).\n", c->name);

	/* Some nice norms */
	p->fr.datalen = 0;
	p->fr.data =  NULL;
	p->fr.src = type;
	p->fr.offset = 0;
	p->fr.mallocd=0;

	if (p->except_struct.type == OH323EXC_USER_INPUT_TONE) {
		p->fr.frametype = AST_FRAME_DTMF;
		p->fr.subclass = p->except_struct.data[0];
		if (option_debug)
			ast_log(LOG_DEBUG, "%s: Got digit %c.\n", c->name, p->fr.subclass);
	} else {
		p->fr.frametype = AST_FRAME_NULL;
		p->fr.subclass = 0;
	}

	/* Clear the exception flag of the channel */
	/* (this is done by Asterisk) */
	/*c->exception = 0;*/

	return(&(p->fr));
}

/**
 * Callback function to activate the exception handling mechanism
 * of ASTERISK. 
 * XXX Possibly, not the best solution. XXX
 */
static int activate_exception(unsigned int call_ref, int type, char *data)
{
	int i;

	if (option_debug)
		ast_log(LOG_DEBUG, "In activate_exception.\n");

	ast_pthread_mutex_lock(&oh323_tab_lock);

	/* Search the already allocated private structures for a */
	/* matching call reference number.*/
	for (i=0; i<config_options.totalNum; i++)
		if (oh323_tab[i] != NULL)
			if (oh323_tab[i]->cd.call_reference == call_ref)
				break;

	/* No match found */
	if ((i<0) || (i>=config_options.totalNum)) {
		if (option_debug)
			ast_log(LOG_WARNING, "Call reference not found.\n");
		ast_pthread_mutex_unlock(&oh323_tab_lock);
		return(-1);
	}

	/* Do we have an owner? */
	if (oh323_tab[i]->owner == NULL) {
		ast_log(LOG_ERROR, "OpenH323 channel with no owner!\n");
		ast_pthread_mutex_unlock(&oh323_tab_lock);
		return(-1);
	}

	/* Fill the exception structure */
	if (type == OH323EXC_USER_INPUT_TONE) {
		(oh323_tab[i]->except_struct).type = type;
		(oh323_tab[i]->except_struct).data[0] = data[0];
	} else {
		ast_log(LOG_ERROR, "Unknown type of exception (%d)!\n", type);
		ast_pthread_mutex_unlock(&oh323_tab_lock);
		return(-1);
	}

	/* Set the exception flag. */
	(oh323_tab[i]->owner)->exception = 1;

	ast_pthread_mutex_unlock(&oh323_tab_lock);
	return(0);
}

/**
 * Indicate condition.
 * Return -1 on error, 0 on success.
 */
static int oh323_indicate(struct ast_channel *c, int condition)
{
	if (option_debug)
		ast_log(LOG_DEBUG, "In oh323_indicate (%s, cond=%d).\n", c->name, condition);

	return(0);
}

/**
 * Send (play) the specified digit to the channel.
 * Return -1 on error, 0 on success.
 */
static int oh323_digit(struct ast_channel *c, char digit)
{
	struct chan_oh323_pvt *pvt = c->pvt->pvt;

	if (option_debug)
		ast_log(LOG_DEBUG, "In oh323_digit (%s, dig=%c).\n", c->name, digit);

	if (c->_state != AST_STATE_UP) {
		ast_log(LOG_WARNING, "%s: Channel is not UP!\n", c->name);
		return(-1);
	}
	h323_send_tone((pvt->cd).call_token, digit);
	return(0);
}

/**
 * Make a call over the specified channel to the specified 
 * destination. This function will parse the destination string
 * and determine the address-number to call.
 * Return -1 on error, 0 on success.
 */
static int oh323_call(struct ast_channel *c, char *dest, int timeout)
{
	struct chan_oh323_pvt *pvt = c->pvt->pvt;
	char called_addr[AST_OH323_MAX_CALLED_ADDR] = {0};
	int retval, i;

	if (option_debug)
		ast_log(LOG_DEBUG, "In oh323_call (%s, dest=%s, timeout=%d).\n", 
						c->name, dest, timeout);

	if ((c->_state != AST_STATE_DOWN) && (c->_state != AST_STATE_RESERVED)) {
		ast_log(LOG_WARNING, "%s: Channel is already in use?\n", c->name);
		return(-1);
	}

	/* Set CallerID on outgoing call */
	if (c->callerid) {
		if (option_debug)
			ast_log(LOG_DEBUG, "%s: Generating CallerID %s\n", c->name, c->callerid);
		if (pvt->call_opt.cid) {
			free(pvt->call_opt.cid);
			pvt->call_opt.cid = NULL;
		}
		pvt->call_opt.cid = (char*)malloc(256);
		if (pvt->call_opt.cid == NULL) {
			ast_log(LOG_ERROR, "Not enough memory.\n");
			return(-1);
		}
		memset(pvt->call_opt.cid, 0, 256);
		strncpy(pvt->call_opt.cid, c->callerid, 255);
	} else {
		if (option_debug)
			ast_log(LOG_DEBUG, "%s: No CallerID info.\n", c->name);
		if (pvt->call_opt.cid) {
			free(pvt->call_opt.cid);
			pvt->call_opt.cid = NULL;
		}
	}

	/* Set other call options */
	pvt->call_opt.connectPort = config_options.connectPort;
	pvt->call_opt.cap = oh323_format2codec(c->nativeformats);
	if (pvt->call_opt.cap == 0) {
		ast_log(LOG_ERROR, "%s: Unsupported format (%d)!\n", 
						c->name, c->nativeformats);
		return(-1);
	}

	/* Clear the call token */
	if ((pvt->cd).call_token == NULL)
		(pvt->cd).call_token = (char *)malloc(AST_OH323_MAX_CALL_TOK);
	if (pvt->cd.call_token == NULL) {
		ast_log(LOG_ERROR, "Not enough memory.\n");
		return(-1);
	}
	memset((char *)(pvt->cd).call_token, 0, AST_OH323_MAX_CALL_TOK);

	ast_pthread_mutex_lock(&oh323_tab_lock);
	pvt->i_state = OH323_STATE_INIT;

	/* Build the address to call. */
	memset(called_addr, 0, AST_OH323_MAX_CALLED_ADDR);
	memcpy(called_addr, dest, strlen(dest));

	/* Make the call */
	retval = h323_make_call(called_addr, &(pvt->cd), pvt->call_opt);

	if (pvt->call_opt.cid) {
		free(pvt->call_opt.cid);
		pvt->call_opt.cid = NULL;
	}
	if (retval != CALL_START_OK) {
		ast_setstate(c, AST_STATE_DOWN);
		ast_pthread_mutex_unlock(&oh323_tab_lock);
		return(-1);
	}
	/* Update statistics */
	ast_pthread_mutex_lock(&oh323_stats_lock);
	++oh323_stats.setup_sent;
	ast_pthread_mutex_unlock(&oh323_stats_lock);

	ast_pthread_mutex_unlock(&oh323_tab_lock);
	/* Wait for a change in the state of the call */
	/* XXX Hardcoded 40 seconds timeout XXX */
	for (i=0; i<400; i++) {
		ast_pthread_mutex_lock(&oh323_tab_lock);
		if (pvt->i_state != OH323_STATE_INIT)
			break;
		ast_pthread_mutex_unlock(&oh323_tab_lock);
		usleep(100*1000);	/* 100 msec delay */
	}
	/* There was a timeout, quit with failure */
	if ((i<0) || (i>=400)) {
		ast_log(LOG_ERROR, "%s: Failed to make call to %s (timeout).\n", 
						c->name, called_addr);
		ast_setstate(c, AST_STATE_DOWN);
		if (pvt->owner != NULL) {
			char shutdown_buf[100] = {0};

			pvt->alreadygone = 1;
			/* XXX Force OpenH323 channel to hangup */
			ast_queue_hangup(pvt->owner, 0);
			write(pvt->player_fd[1], shutdown_buf, 100);
		}
		ast_pthread_mutex_unlock(&oh323_tab_lock);
		return(-1);
	}
	/* The call cleared (error), quit with failure */
	if (pvt->i_state == OH323_STATE_CLEARED) {
		ast_log(LOG_ERROR, "%s: Could not call %s.\n", c->name, called_addr);
		ast_setstate(c, AST_STATE_DOWN);
		if (pvt->owner != NULL) {
			char shutdown_buf[100] = {0};

			pvt->alreadygone = 1;
			/* XXX Force OpenH323 channel to hangup */
			ast_queue_hangup(pvt->owner, 0);
			write(pvt->player_fd[1], shutdown_buf, 100);
		}
		ast_pthread_mutex_unlock(&oh323_tab_lock);
		return(-1);
	}

	/* The call initiated successfully (remote endpoint is ringing) */
	if (option_debug)
		ast_log(LOG_DEBUG, "%s: Call initiated successfully.\n", c->name);
	/* Set the state of the channel and fix the name */
	ast_setstate(c, AST_STATE_RINGING);
	memset(c->name, 0, sizeof(c->name));
	snprintf(c->name, sizeof(c->name)-1, "H323:%d", pvt->cd.call_reference);
	ast_pthread_mutex_unlock(&oh323_tab_lock);

	return(0);
}

/**
 * Send a hungup to the specified channel. This function
 * will terminate the active call over this channel, will
 * decrease the usage counter and will free allocated resources.
 * It can not fail !!!
 */
static int oh323_hangup(struct ast_channel *c) 
{
	struct chan_oh323_pvt *pvt = c->pvt->pvt;
	int i;

	if (option_debug)
		ast_log(LOG_DEBUG, "In oh323_hangup (%s).\n", c->name);

	ast_pthread_mutex_lock(&oh323_tab_lock);
	if (pvt) {
		ast_setstate(c, AST_STATE_DOWN);
		/*ast_softhangup(c, 1);*/
		ast_queue_hangup(c, 0);

		/* Send the hangup */
		if (!(pvt->alreadygone)) {
			if (option_debug)
				ast_log(LOG_DEBUG, "%s: Call token is %s.\n", c->name, 
								(pvt->cd).call_token);
			if (h323_clear_call((pvt->cd).call_token) != CALL_END_OK) 
				ast_log(LOG_WARNING, "%s: Failed to hangup channel.\n", c->name);
		}

		/* Search the already allocated private structures for a */
		/* matching call reference number.*/
		for (i=0; i<config_options.totalNum; i++)
			if (oh323_tab[i] != NULL)
				if (oh323_tab[i]->cd.call_reference == pvt->cd.call_reference)
					if (strcmp(oh323_tab[i]->cd.call_token, pvt->cd.call_token) == 0)
						break;

		/* No match found ?!?!?! */
		if ((i<0) || (i>=config_options.totalNum)) {
			if (option_debug)
				ast_log(LOG_WARNING, "%s: Call reference not found!\n", c->name);
		} else {
			if (option_debug)
				ast_log(LOG_DEBUG, "%s: Call reference match found for %d in %d (%d bytes).\n", 
							c->name, pvt->cd.call_reference, i, sizeof(*(oh323_tab[i])));

			/* Deallocate the resources */
			oh323_destroy(i);
			free(oh323_tab[i]);
			oh323_tab[i] = NULL;
		}

		c->pvt->pvt = NULL;

		/* Update usage counter */
		ast_pthread_mutex_lock(&usecnt_lock);
		usecnt--;
		if (usecnt < 0) 
	 		ast_log(LOG_WARNING, "Usecnt < 0???\n");
		ast_pthread_mutex_unlock(&usecnt_lock);
		ast_update_use_count();
		if (option_verbose > 2) 
			ast_verbose( VERBOSE_PREFIX_3 "Hungup '%s'\n", c->name);
	}
	ast_pthread_mutex_unlock(&oh323_tab_lock);
	
	return(0);
}

/**
 * Read data from the network, encapsulate them in a frame
 * and return it to Asterisk PBX.
 * Return NULL on error, a frame pointer on success.
 */
static struct ast_frame *oh323_read(struct ast_channel *c)
{
	struct chan_oh323_pvt *p = c->pvt->pvt;
	int res, read_size;
	struct ast_frame *f;

	/*ast_log(LOG_DEBUG, "In oh323_read, channel %s.\n", c->name);*/

	/* If it's already gone, just return */
	if ((p->alreadygone)||(c->_state == AST_STATE_DOWN)) {
		ast_log(LOG_DEBUG, "%s: Channel is DOWN.\n", c->name);
		return(NULL);
	}

	/* H323 channels have been established? */
	/************
	if ((p->i_state == OH323_STATE_ESTABLISHED)&&(c->_state == AST_STATE_RINGING)) {
		ast_log(LOG_DEBUG, "H323 state is ESTABLISHED, going UP (%s).\n", c->name);
		p->fr.frametype = AST_FRAME_CONTROL;
		p->fr.subclass = AST_CONTROL_ANSWER;
		return &(p->fr);
	}
	************/

	/* If the channel is in RINGING state and we get something
	 * from the network, then build a control ANSWER frame and
	 * return it to ASTERISK. This way, we trigger ASTERISK to
	 * answer the channel and establish the call.
	 */
	/*if ((c->_state == AST_STATE_RINGING)||(c->_state == AST_STATE_RING)) {*/
	if ((p->i_state == OH323_STATE_ESTABLISHED)&&(c->_state == AST_STATE_RINGING)) {
		if (option_debug)
			ast_log(LOG_DEBUG, "%s: Channel is going UP.\n", c->name);
		p->fr.frametype = AST_FRAME_CONTROL;
		p->fr.subclass = AST_CONTROL_ANSWER;
		return &(p->fr);
	}

	/*if (p->i_state != OH323_STATE_ESTABLISHED) {
		ast_log(LOG_DEBUG, "%s: H323 state is not ESTABLISHED (%d).\n", 
							c->name, p->i_state);
		p->fr.frametype = AST_FRAME_NULL;
		p->fr.subclass = 0;
		return &(p->fr);
	}*/

	/* Find out how much data we are going to read */
	if (p->rx_buf_size <= 0) {
		read_size = OH323_MAX_BUF;
	} else {
		if (p->rx_buf_size > OH323_MAX_BUF) {
			ast_log(LOG_WARNING, "%s: Requested read buffer size is too long (%d)!\n",
							c->name, p->rx_buf_size);
			ast_log(LOG_WARNING, "%s: Truncating to %d.\n", c->name, OH323_MAX_BUF);
			read_size = OH323_MAX_BUF;
		} else
			read_size = p->rx_buf_size;
	}

	/* Try to read some data... */
	CHECK_BLOCKING(c);
	res = read(p->player_fd[0], (char *)(p->buf), read_size);
	c->blocking = 0;
	if (res <= 0) {
		/* This will hangup the channel. */
		if (res < 0) {
			ast_log(LOG_ERROR, "%s: Error reading from channel (%s).\n", 
							c->name, strerror(errno));
		} else {
			if (option_debug)
				ast_log(LOG_WARNING, "%s: Zero-size read on channel.\n", c->name);
		}
		p->alreadygone = 1;
		return(NULL);
	}
	if (oh323_verbose_debug)
		ast_log(LOG_DEBUG, "%s: Read %d bytes.\n", c->name, res);

	/* Do specific stuff for each format type */
	switch (p->capability) {
		case AST_FORMAT_ULAW:
		case AST_FORMAT_ALAW:
			break;

		case AST_FORMAT_GSM:
			if (res % 33 != 0) {
				ast_log(LOG_WARNING, "%s: Invalid frame size for GSM (%d bytes).\n", 
						c->name, res);
				p->fr.frametype = AST_FRAME_NULL;
				p->fr.subclass = 0;
				return &(p->fr);
			}
			break;

		case AST_FORMAT_G729A:
			if (res % 10 != 0) {
				ast_log(LOG_WARNING, "%s: Invalid frame size for G.729A (%d bytes).\n", 
						c->name, res);
				p->fr.frametype = AST_FRAME_NULL;
				p->fr.subclass = 0;
				return &(p->fr);
			}
			break;

		case AST_FORMAT_G723_1:
			if ((res % 24 != 0)&&(res % 20 != 0)) {
				/*ast_log(LOG_WARNING, "%s: Invalid frame size for G.723.1 (%d bytes).\n", 
						c->name, res);*/
				p->fr.frametype = AST_FRAME_NULL;
				p->fr.subclass = 0;
				return &(p->fr);
			}
			break;

		default:
			/* Do nothing */
			break;
	}

	if (p->i_state != OH323_STATE_ESTABLISHED) {
		if (oh323_verbose_debug)
			ast_log(LOG_DEBUG, "%s: H323 state is not ESTABLISHED (%d).\n", 
								c->name, p->i_state);
		p->fr.frametype = AST_FRAME_NULL;
		p->fr.subclass = 0;
		return &(p->fr);
	}

	/* Save the data in a ASTERISK frame struct */
	p->fr.mallocd = 0;
	p->fr.src = type;
	p->fr.data = p->buf;
	p->fr.datalen = res;
	p->fr.frametype = AST_FRAME_VOICE;
	p->fr.subclass = p->capability;
	p->fr.offset = 0;
	f = &(p->fr);

	/* Do in-band DTMF detection, if enabled */
	if ((p->capability==AST_FORMAT_ALAW)||(p->capability==AST_FORMAT_ULAW)) {
		if (config_options.inBandDTMF) {
			f = ast_dsp_process(p->owner, p->vad, f, 0);
			if (f->frametype == AST_FRAME_DTMF)
				if (option_debug)
					ast_log(LOG_DEBUG, "%s: Got in-band digit %c.\n", c->name, f->subclass);
		}
	}

	return(f);
}

/**
 * Write the data buffer of the frame to the network.
 * Return -1 on error, 0 on success.
 */
static int oh323_write(struct ast_channel *c, struct ast_frame *f)
{
	struct chan_oh323_pvt	*i = c->pvt->pvt;
	struct ast_frame		*ftmp;
	int						res;

	/*ast_log(LOG_DEBUG, "In oh323_write.\n");*/

	/* H323 channels have been established? */
	if (i->i_state != OH323_STATE_ESTABLISHED) {
		if (option_debug)
			ast_log(LOG_DEBUG, "%s: H323 state is not ESTABLISHED (%d).\n", 
							c->name, i->i_state);
		return(0);
	}

	if (i->tx_smooth == NULL) {
		i->tx_smooth = ast_smoother_new(i->tx_buf_size);
		if (i->tx_smooth == NULL) {
			ast_log(LOG_ERROR, "%s: Failed to create smoother.\n", c->name);
			return(-1);
		}
		if (option_debug)
			ast_log(LOG_DEBUG, "%s: Created smoother (size=%d).\n", 
						c->name, i->tx_buf_size);
	}

	/* If it's already gone, return failure */
	if ((i->alreadygone)||(c->_state == AST_STATE_DOWN)) {
		if (option_debug)
			ast_log(LOG_DEBUG, "%s: Channel is DOWN.\n", c->name);
		return(-1);
	}

	if (c->_state != AST_STATE_UP) {
		if (option_debug)
			ast_log(LOG_DEBUG, "%s: Channel is not UP yet.\n", c->name);
		return(0);
	}

	/* Don't waste bandwidth sending null frames */
	if (f->frametype == AST_FRAME_NULL) {
		if (option_debug)
			ast_log(LOG_DEBUG, "%s: Ignoring NULL frame.\n", c->name);
		return(0);
	}

	/* Check for a proper frame type. */
	if (f->frametype != AST_FRAME_VOICE) {
		ast_log(LOG_WARNING, "%s: Don't know what to do with frame type '%d'.\n", 
						c->name, f->frametype);
		ast_frfree(f);
		return(-1);
	}

	/* Do specific stuff for each codec type */
	switch (f->subclass) {
		case AST_FORMAT_ULAW:
		case AST_FORMAT_ALAW:
			break;

		case AST_FORMAT_GSM:
			if (f->datalen % 33 != 0) {
				ast_log(LOG_WARNING, "%s: Invalid frame size for GSM (%d bytes).\n", 
						c->name, f->datalen);
				return(0);
			}
			break;

		case AST_FORMAT_G729A:
			if (f->datalen % 10 != 0) {
				ast_log(LOG_WARNING, "%s: Invalid frame size for G.729A (%d bytes).\n", 
						c->name, f->datalen);
				return(0);
			}
			break;

		case AST_FORMAT_G723_1:
			if ((f->datalen % 24 != 0)&&(f->datalen % 20 != 0)) {
				ast_log(LOG_WARNING, "%s: Invalid frame size for G.723.1 (%d bytes).\n", 
						c->name, f->datalen);
				return(0);
			}
			break;
			
		default:
			/* Do nothing */
			break;
	}

	if (ast_smoother_feed(i->tx_smooth, f) != 0) {
		ast_log(LOG_ERROR, "%s: Failed to fill smoother.\n", c->name);
		return(-1);
	}
	ftmp = ast_smoother_read(i->tx_smooth);
	while (ftmp != NULL) {
		if (oh323_verbose_debug)
			ast_log(LOG_DEBUG, "%s: Writing %d bytes...\n", c->name, ftmp->datalen);
		if ((res = write(i->recorder_fd[0], (char *)(ftmp->data), ftmp->datalen)) < 0) {
			if (option_debug)
				ast_log(LOG_WARNING, "%s: Unable to write to network (errno=%d).\n", 
							c->name, errno);
			return(0);
		}
		if (oh323_verbose_debug)
			ast_log(LOG_DEBUG, "%s: Wrote %d bytes...\n", c->name, ftmp->datalen);
		ftmp = ast_smoother_read(i->tx_smooth);
	}

	return(0);
}

static int oh323_answer(struct ast_channel *c)
{
	struct chan_oh323_pvt *pvt = c->pvt->pvt;
	int	i;

	if (option_debug)
		ast_log(LOG_DEBUG, "In oh323_answer (%s).\n", c->name);

	if (pvt->i_state == OH323_STATE_ESTABLISHED) {
		ast_log(LOG_DEBUG, "%s: Channel is established???\n", c->name);
		return(0);
	}
	if (c->_state == AST_STATE_UP) {
		ast_log(LOG_DEBUG, "%s: Channel is UP???\n", c->name);
		return(0);
	}

	/*ast_pthread_mutex_lock(&oh323_tab_lock);*/
	if (h323_answer_call((pvt->cd).call_token) != CALL_ANS_OK) {
		ast_log(LOG_WARNING, "%s: Failed to answer call with token %s.\n",
						c->name, (pvt->cd).call_token);
		/*ast_pthread_mutex_unlock(&oh323_tab_lock);*/
		return(-1);
	}

	/*ast_pthread_mutex_unlock(&oh323_tab_lock);*/

	/* Wait for the channel to be established */
	/* XXX Hardcoded 20 seconds timeout XXX */
	for (i=0; i<200; i++) {
		ast_pthread_mutex_lock(&oh323_tab_lock);
		if (pvt->i_state == OH323_STATE_ESTABLISHED)
			break;
		ast_pthread_mutex_unlock(&oh323_tab_lock);
		usleep(100*1000);	/* 100 msec delay */
	}
	/* There was a timeout, quit with failure */
	if ((i<0) || (i>=200)) {
		ast_log(LOG_ERROR, "%s: Failed to answer call with token %s (timeout).\n", 
						c->name, (pvt->cd).call_token);
		ast_setstate(c, AST_STATE_DOWN);
		ast_pthread_mutex_unlock(&oh323_tab_lock);
		return(-1);
	}

	/* The call was answered successfully */
	if (option_debug)
		ast_log(LOG_DEBUG, "%s: Call answered.\n", c->name);
	/* XXX ast_setstate(c, AST_STATE_UP);*/
	ast_pthread_mutex_unlock(&oh323_tab_lock);

	return(0);
}

static int oh323_fixup(struct ast_channel *oldchan, struct ast_channel *newchan)
{
	struct chan_oh323_pvt *p = newchan->pvt->pvt;
	
	ast_pthread_mutex_lock(&oh323_tab_lock);
	if (p->owner != oldchan) {
		ast_log(LOG_WARNING, "old channel wasn't %p but was %p\n", oldchan, p->owner);
		ast_pthread_mutex_unlock(&oh323_tab_lock);
		return(-1);
	}
	p->owner = newchan;
	ast_pthread_mutex_unlock(&oh323_tab_lock);
	return(0);
}

/**
 * Allocate and initialize a "chan_oh323_pvt"
 * private structure.
 * Return NULL on error, the pointer on success.
 */
static struct chan_oh323_pvt *new_oh323(int index)
{
	struct chan_oh323_pvt	*tmp;
	int						flags, i;

	tmp = malloc(sizeof(struct chan_oh323_pvt));
	if (tmp) {
		memset(tmp, 0, sizeof(struct chan_oh323_pvt));

		/* Create socketpairs */
		snprintf(tmp->playername, sizeof(tmp->playername), "%s", 
						build_fifo_name(PLAYER, index));
		snprintf(tmp->recordername, sizeof(tmp->recordername), "%s", 
						build_fifo_name(RECORDER, index));
		if (socketpair(AF_LOCAL, SOCK_STREAM, 0, tmp->player_fd) != 0) {
			ast_log(LOG_ERROR, "Failed to create socketpair for player(%d, %s).\n",
							errno, strerror(errno));
			free(tmp);
			return NULL;
		}
		if (shutdown(tmp->player_fd[1], SHUT_RD) != 0) {
			ast_log(LOG_ERROR, "Failed to configure player socket.\n");
			free(tmp);
			return NULL;
		}
		if (socketpair(AF_LOCAL, SOCK_STREAM, 0, tmp->recorder_fd) != 0) {
			ast_log(LOG_ERROR, "Failed to create socketpair for recorder(%d, %s).\n",
							errno, strerror(errno));
			free(tmp);
			return NULL;
		}
		if (shutdown(tmp->recorder_fd[1], SHUT_WR) != 0) {
			ast_log(LOG_ERROR, "Failed to configure recorder socket.\n");
			free(tmp);
			return NULL;
		}
		/* Set all fds in non-blocking mode */
		for (i=0; i<2; i++) {
			flags = fcntl(tmp->player_fd[i], F_GETFL);
			if (fcntl(tmp->player_fd[i], F_SETFL, flags | O_NONBLOCK) < 0) {
				ast_log(LOG_ERROR, "Failed to configure player socket.\n");
				free(tmp);
				return NULL;
			}
		}
		for (i=0; i<2; i++) {
			flags = fcntl(tmp->recorder_fd[i], F_GETFL);
			if (fcntl(tmp->recorder_fd[i], F_SETFL, flags | O_NONBLOCK) < 0) {
				ast_log(LOG_ERROR, "Failed to configure recorder socket.\n");
				free(tmp);
				return NULL;
			}
		}

		tmp->direction = NONE;
		tmp->alreadygone = 0;
		tmp->from_remote = 0;
		tmp->tx_smooth = NULL;
		tmp->i_state = OH323_STATE_NULL;
	}
	return tmp;
}


/**
 * Destroy allocated resources of an existing "chan_oh323_pvt"
 * private structure. Do not free the allocated memory.
 * Returns nothing.
 */
static void oh323_destroy(int index)
{
	struct chan_oh323_pvt *pvt = oh323_tab[index];

	if (option_debug)
		ast_log(LOG_DEBUG, "Releasing allocated resources (index=%d).\n", index);

	/* Close open sockets */
	clear_call_details(&(pvt->cd));
	close(pvt->player_fd[0]);
	close(pvt->recorder_fd[0]);
	if (pvt->i_state != OH323_STATE_ESTABLISHED) {
		close(pvt->player_fd[1]);
		close(pvt->recorder_fd[1]);
	}
	if ((config_options.inBandDTMF)&&(pvt->vad != NULL))
		ast_dsp_free(pvt->vad);
	pvt->i_state = OH323_STATE_NULL;
	pvt->direction = NONE;
	pvt->alreadygone = 1;
	pvt->owner = NULL;
}

/**
 * Allocate and initialize a "ast_channel" (ASTERISK channel)
 * structure.
 */
static struct ast_channel *ast_oh323_new(struct chan_oh323_pvt *i, int state)
{
	struct ast_channel *tmp;
	int callerid_len;
	char *context;

	if (i == NULL) {
		ast_log(LOG_ERROR, "Private structure is NULL!\n");
		return NULL;
	}

	tmp = ast_channel_alloc(0);
	if (tmp) {
		snprintf(tmp->name, sizeof(tmp->name)-1, "H323:%d", i->cd.call_reference);
		tmp->type = type;

		if (config_options.inBandDTMF) {
			i->vad = ast_dsp_new();
			ast_dsp_set_features(i->vad, DSP_FEATURE_DTMF_DETECT);
		}
		tmp->fds[0] = i->player_fd[0];

		tmp->nativeformats = oh323_capability;
		tmp->pvt->pvt = i;
		/* Register the OpenH323 channel's functions. */
		tmp->pvt->call = oh323_call;
		tmp->pvt->hangup = oh323_hangup;
		tmp->pvt->answer = oh323_answer;
		tmp->pvt->read = oh323_read;
		tmp->pvt->write = oh323_write;
		tmp->pvt->send_digit = oh323_digit;
		tmp->pvt->exception = oh323_exception;
		tmp->pvt->indicate = oh323_indicate;
		tmp->pvt->fixup = oh323_fixup;
		i->owner = tmp;
		ast_setstate(tmp, state);
		tmp->blocking = 0;

		/* Set initial context and extension */
		/* First, search for a matching alias, then for a matching prefix */
		/* or return the default context. */
		/* The extension is the same as the called E.164 address of the call, */
		/* or the default extension, if this cannot be determined */
		strncpy(tmp->context, config_options.context, sizeof(tmp->context)-1);
		if (i->from_remote) {
			if (i->cd.call_dest_e164 != NULL) {
				if (strlen(i->cd.call_dest_e164) > 0) {
					if (!context_from_alias((char*)i->cd.call_dest_e164, &context))
						strncpy(tmp->context, context, sizeof(tmp->context)-1);
					else if (!context_from_prefix((char*)i->cd.call_dest_e164, &context))
						strncpy(tmp->context, context, sizeof(tmp->context)-1);
					else
						strncpy(tmp->context, config_options.context, sizeof(tmp->context)-1);
					strncpy(tmp->exten, i->cd.call_dest_e164, sizeof(tmp->exten)-1);
				} 
			}
		}

		/* Set caller ID */
		if (i->cd.call_source_e164 != NULL) {
			if (strlen(i->cd.call_source_e164) > 0) {
				callerid_len = strlen(i->cd.call_source_e164)+5;
				tmp->callerid = (char *)malloc(callerid_len);
				if (tmp->callerid != NULL) {
					memset(tmp->callerid, 0, callerid_len);
					memcpy(tmp->callerid, i->cd.call_source_e164, 
									strlen(i->cd.call_source_e164));
				}
			}
		}
		if (option_debug) {
			ast_log(LOG_DEBUG, "Context is %s, extension is %s.\n", 
							tmp->context, tmp->exten);
			if (tmp->callerid != NULL)
				ast_log(LOG_DEBUG, "Caller ID is %s.\n", tmp->callerid);
		}

		/* Set channel's AMA flags and account code */
		if (config_options.amaFlags >= 0)
			tmp->amaflags = config_options.amaFlags;
		if (strlen(config_options.accountCode))
			strncpy(tmp->accountcode, config_options.accountCode, sizeof(tmp->accountcode)-1);

		/* Update use count */
		ast_pthread_mutex_lock(&usecnt_lock);
		usecnt++;
		ast_pthread_mutex_unlock(&usecnt_lock);
		ast_update_use_count();

		if (state != AST_STATE_DOWN) {
			if (ast_pbx_start(tmp)) {
				ast_log(LOG_WARNING, "Unable to start PBX on %s.\n", tmp->name);
				ast_hangup(tmp);
				tmp = NULL;
			}
		}
	}
	return tmp;
}

static struct ast_channel *oh323_request(char *type, int format, void *data)
{
	int					i=0, count, oldformat;
	struct ast_channel	*c;

	if (option_debug) {
		ast_log(LOG_DEBUG, "In oh323_request.\n");
		ast_log(LOG_DEBUG, "type=%s, format=%d, data=%s.\n", type, format, (char *)data);
	}

	/* Check the format requested */
	oldformat = format;
	format &= oh323_capability;
	if (!format) {
		ast_log(LOG_ERROR, "Asked to get a channel of unsupported format '%d'\n", format);
		return NULL;
	}

	ast_pthread_mutex_lock(&oh323_tab_lock);

	/* Check the number of active oubound H.323 calls */
	count = 0;
	for (i=0; i<config_options.totalNum; i++)
		if (oh323_tab[i] != NULL)
			if (oh323_tab[i]->from_remote == 0)
				++count;
	if (count >= config_options.outboundMax) {
		ast_log(LOG_WARNING, "Blocking outbound H.323 call due to call-limit violation.\n");

		/* Update statistics */
		ast_pthread_mutex_lock(&oh323_stats_lock);
		++oh323_stats.block_outcall;
		ast_pthread_mutex_unlock(&oh323_stats_lock);

		ast_pthread_mutex_unlock(&oh323_tab_lock);
		return NULL;
	}

	/* Search for a free entry in the private structure table. */
	for (i=0; i<config_options.totalNum; i++)
		if (oh323_tab[i] == NULL)
			break;

	/* No free entry */
	if ((i<0) || (i>=config_options.totalNum)) {
		ast_log(LOG_ERROR, "Unable to accept more calls.\n");
		ast_pthread_mutex_unlock(&oh323_tab_lock);
		return NULL;
	}

	oh323_tab[i] = new_oh323(i);
	if (oh323_tab[i]) {
		if (option_debug)
			ast_log(LOG_DEBUG, "Created new call structure %d (%d bytes).\n", 
						i, sizeof(*(oh323_tab[i])));
		c = ast_oh323_new(oh323_tab[i], AST_STATE_DOWN);
		if (c) {
			oh323_tab[i]->owner = c;

			/* Choose a format we can live with */
			c->nativeformats = format;

			/* XXX Other stuff to do...??? */

			ast_pthread_mutex_unlock(&oh323_tab_lock);
			return c;
		}
	}

	ast_log(LOG_WARNING, "Failed to create new H.323 private structure %d.\n", i);
	ast_pthread_mutex_unlock(&oh323_tab_lock);
	return NULL;
}

/**
 * Copy the call_details structure to a local structure.
 * All the "const char *" variables are malloc'ed with this function.
 * The right cleanup should be done by calling the "clear_call_details".
 * Return -1 on error, 0 on success.
 */
static int copy_call_details(call_details_t *src_cd, call_details_t *dest_cd)
{
	clear_call_details(dest_cd);

	dest_cd->call_reference = src_cd->call_reference;

	/* Clear "call_token" */
	dest_cd->call_token = (char *)malloc(AST_OH323_MAX_CALL_TOK);
	if (dest_cd->call_token == NULL) {
		ast_log(LOG_ERROR, "Not enough memory.\n");
		return -1;
	}
	memset((char *)dest_cd->call_token, 0, AST_OH323_MAX_CALL_TOK);
	/* Clear "call_source_alias" */
	dest_cd->call_source_alias = (char *)malloc(AST_OH323_MAX_CALL_ADDR);
	if (dest_cd->call_source_alias == NULL) {
		ast_log(LOG_ERROR, "Not enough memory.\n");
		free((char *)dest_cd->call_token);
		return -1;
	}
	memset((char *)dest_cd->call_source_alias, 0, AST_OH323_MAX_CALL_ADDR);
	/* Clear "call_dest_alias" */
	dest_cd->call_dest_alias = (char *)malloc(AST_OH323_MAX_CALL_ADDR);
	if (dest_cd->call_dest_alias == NULL) {
		ast_log(LOG_ERROR, "Not enough memory.\n");
		free((char *)dest_cd->call_token);
		free((char *)dest_cd->call_source_alias);
		return -1;
	}
	memset((char *)dest_cd->call_dest_alias, 0, AST_OH323_MAX_CALL_ADDR);
	/* Clear "call_source_e164" */
	dest_cd->call_source_e164 = (char *)malloc(AST_OH323_MAX_CALL_ADDR);
	if (dest_cd->call_source_e164 == NULL) {
		ast_log(LOG_ERROR, "Not enough memory.\n");
		free((char *)dest_cd->call_token);
		free((char *)dest_cd->call_source_alias);
		free((char *)dest_cd->call_dest_alias);
		return -1;
	}
	memset((char *)dest_cd->call_source_e164, 0, AST_OH323_MAX_CALL_ADDR);
	/* Clear "call_dest_e164" */
	dest_cd->call_dest_e164 = (char *)malloc(AST_OH323_MAX_CALL_ADDR);
	if (dest_cd->call_dest_e164 == NULL) {
		ast_log(LOG_ERROR, "Not enough memory.\n");
		free((char *)dest_cd->call_token);
		free((char *)dest_cd->call_source_alias);
		free((char *)dest_cd->call_dest_alias);
		free((char *)dest_cd->call_source_e164);
		return -1;
	}
	memset((char *)dest_cd->call_dest_e164, 0, AST_OH323_MAX_CALL_ADDR);

	/* Copy bytes */
	memcpy((char *)dest_cd->call_token, src_cd->call_token, strlen(src_cd->call_token));
	memcpy((char *)dest_cd->call_source_alias, src_cd->call_source_alias, strlen(src_cd->call_source_alias));
	memcpy((char *)dest_cd->call_dest_alias, src_cd->call_dest_alias, strlen(src_cd->call_dest_alias));
	memcpy((char *)dest_cd->call_source_e164, src_cd->call_source_e164, strlen(src_cd->call_source_e164));
	memcpy((char *)dest_cd->call_dest_e164, src_cd->call_dest_e164, strlen(src_cd->call_dest_e164));

	if (option_debug) {
		ast_log(LOG_DEBUG, "--- CALL DETAILS ---\n");
		ast_log(LOG_DEBUG, "call_token = %s\n", dest_cd->call_token);
		ast_log(LOG_DEBUG, "call_source_alias = %s\n", dest_cd->call_source_alias);
		ast_log(LOG_DEBUG, "call_dest_alias = %s\n", dest_cd->call_dest_alias);
		ast_log(LOG_DEBUG, "call_source_e164 = %s\n", dest_cd->call_source_e164);
		ast_log(LOG_DEBUG, "call_dest_e164 = %s\n", dest_cd->call_dest_e164);
	}
	
	return(0);
}

/**
 * Cleanup all the "const char *" variables of the call_details structure.
 * Returns nothing.
 */
static void clear_call_details(call_details_t *cd)
{
	if (cd->call_token != NULL) {
		free((char *)cd->call_token);
		cd->call_token = NULL;
	}
	if (cd->call_source_alias != NULL) {
		free((char *)cd->call_source_alias);
		cd->call_source_alias = NULL;
	}
	if (cd->call_dest_alias != NULL) {
		free((char *)cd->call_dest_alias);
		cd->call_dest_alias = NULL;
	}
	if (cd->call_source_e164 != NULL) {
		free((char *)cd->call_source_e164);
		cd->call_source_e164 = NULL;
	}
	if (cd->call_dest_e164 != NULL) {
		free((char *)cd->call_dest_e164);
		cd->call_dest_e164 = NULL;
	}
}


/**
 * Determine the ASTERISK format of this channel by the specified codec.
 * Returns non-zero in success, 0 in case of unsupported codec.
 */
static int oh323_codec2format(int codec)
{
	switch (codec) {
		case G711U:
			/*return(AST_FORMAT_SLINEAR);*/
			return(AST_FORMAT_ULAW);
		case G711A:
			/*return(AST_FORMAT_SLINEAR);*/
			return(AST_FORMAT_ALAW);
		case G7231: /* fall-through */
		case G72316K3: /* fall-through */
		case G72315K3: /* fall-through */
		case G7231A6K3: /* fall-through */
		case G7231A5K3:
			return(AST_FORMAT_G723_1);
		case GSM0610:
			return(AST_FORMAT_GSM);
		case LPC10:
			return(AST_FORMAT_SLINEAR);
		case G729:
		case G729A:
		case G729B:
		case G729AB:
			return(AST_FORMAT_G729A);
		case MSGSM: /* fall-through */
		default:
			return(0);
	}
}

static char *oh323_capability2str(int capability, char **res)
{
	*res = (char *)malloc(1024);
	if (*res == NULL)
		return(NULL);
	memset(*res, 0, 1024);

	if (capability & AST_FORMAT_G723_1)
		strcat(*res, "G.723.1 ");
	if (capability & AST_FORMAT_GSM)
		strcat(*res, "GSM ");
	if (capability & AST_FORMAT_ULAW)
		strcat(*res, "G.711U ");
	if (capability & AST_FORMAT_ALAW)
		strcat(*res, "G.711A ");
	if (capability & AST_FORMAT_SLINEAR)
		strcat(*res, "SLINEAR ");
	if (capability & AST_FORMAT_G729A)
		strcat(*res, "G.729A ");
	if (capability & AST_FORMAT_LPC10)
		strcat(*res, "LPC10 ");
	return(*res);
}

/**
 * Determine the code type by the specified ASTERISK format.
 * Returns non-zero in success, 0 in case of unsupported codec.
 */
static int oh323_format2codec(int format)
{
	switch (format) {
		case AST_FORMAT_G723_1:
			return(G7231);
		case AST_FORMAT_GSM:
			return(GSM0610);
		case AST_FORMAT_ULAW:
			return(G711U);
		case AST_FORMAT_ALAW:
			return(G711A);
		case AST_FORMAT_SLINEAR:
			return(G711U);
		case AST_FORMAT_G729A:
			return(G729A);
		case AST_FORMAT_LPC10:
			return(G711U);
		default:
			return(0);
	}
}

static char *oh323_format2str(int format)
{
	switch (format) {
		case AST_FORMAT_G723_1:
			return("G.723.1 ");
		case AST_FORMAT_GSM:
			return("GSM ");
		case AST_FORMAT_ULAW:
			return("G.711U ");
		case AST_FORMAT_ALAW:
			return("G.711A ");
		case AST_FORMAT_SLINEAR:
			return("SLINEAR ");
		case AST_FORMAT_G729A:
			return("G.729A ");
		case AST_FORMAT_LPC10:
			return("LPC10 ");
	}
	return(NULL);
}

#if 0
/**
 * Convert H.323 codec numerical values into names.
 */
static char *oh323_codec2str(int val)
{
	switch (val) {
		case G711U:
			return("G711U");
		case G711A:
			return("G711A");
		case G7231:
			return("G7231");
		case G72316K3:
			return("G72316K3");
		case G72315K3:
			return("G72315K3");
		case G7231A6K3:
			return("G7231A6K3");
		case G7231A5K3:
			return("G7231A5K3");
		case G728:
			return("G728");
		case G729:
			return("G729");
		case G729A:
			return("G729A");
		case G729B:
			return("G729B");
		case G729AB:
			return("G729AB");
		case GSM0610:
			return("GSM0610");
		case MSGSM:
			return("MSGSM");
		case LPC10:
			return("LPC10");
	}
	return NULL;
}
#endif

/**
 * Convert H.323 codec names into numerical values.
 */
static int oh323_str2codec(char *name)
{
	if (!strcasecmp(name, "G711U")) 
		return G711U;
	else if (!strcasecmp(name, "G711A"))
		return G711A;
	else if (!strcasecmp(name, "G7231"))
		return G7231;
	else if (!strcasecmp(name, "G72316K3"))
		return G72316K3;
	else if (!strcasecmp(name, "G72315K3"))
		return G72315K3;
	else if (!strcasecmp(name, "G7231A6K3"))
		return G7231A6K3;
	else if (!strcasecmp(name, "G7231A5K3"))
		return G7231A5K3;
	else if (!strcasecmp(name, "G728"))
		return G728;
	else if (!strcasecmp(name, "G729"))
		return G729;
	else if (!strcasecmp(name, "G729A"))
		return G729A;
	else if (!strcasecmp(name, "G729B"))
		return G729B;
	else if (!strcasecmp(name, "G729AB"))
		return G729AB;
	else if (!strcasecmp(name, "GSM0610"))
		return GSM0610;
	else if (!strcasecmp(name, "MSGSM"))
		return MSGSM;
	else if (!strcasecmp(name, "LPC10"))
		return LPC10;
	return 0;
}

/**
 * Convert H.323 channel numerical states into strings.
 */
static char *oh323_state2str(int state)
{
	switch (state) {
		case OH323_STATE_NULL:
			return("NULL");
		case OH323_STATE_INIT:
			return("INIT");
		case OH323_STATE_RING:
			return("RING");
		case OH323_STATE_RINGING:
			return("RINGING");
		case OH323_STATE_ESTABLISHED:
			return("ESTABLISHED");
		case OH323_STATE_CLEARED:
			return("CLEARED");
		default:
			return("UNKNOWN");
	}
}

/**
 * Find under which context the specified "alias" resides.
 * Return the pointer to the context in "context" and 0 on success,
 * -1 on error or no match.
 */
static int context_from_alias(char *alias, char **context)
{
	struct oh323_reginfo *regtmp;
	int i;

	if ((alias == NULL)||(context == NULL)||(strlen(alias) <= 0))
		return(-1);

	regtmp = config_options.regInfo;
	*context = NULL;
	while (regtmp) {
		for (i=0; i<regtmp->alias_num; i++)
			if (!strcasecmp(regtmp->alias[i], alias)) {
				*context = regtmp->context;
				if (option_debug)
					ast_log(LOG_DEBUG, "Routing alias %s in context %s.\n",
									alias, *context);
				return(0);
			}
		regtmp = regtmp->next;
	}
	return(-1);
}

/**
 * Find under which context the specified "prefix" resides.
 * Return the pointer to the context in "context" and 0 on success,
 * -1 on error or no match.
 */
static int context_from_prefix(char *prefix, char **context)
{
	struct oh323_reginfo *regtmp;
	int i;

	if ((prefix == NULL)||(context == NULL)||(strlen(prefix) <= 0))
		return(-1);

	regtmp = config_options.regInfo;
	*context = NULL;
	while (regtmp) {
		for (i=0; i<regtmp->prefix_num; i++)
			if (!strncasecmp(regtmp->prefix[i], prefix, strlen(regtmp->prefix[i]))) {
				*context = regtmp->context;
				if (option_debug)
					ast_log(LOG_DEBUG, "Routing prefix %s in context %s.\n",
									prefix, *context);
				return(0);
			}
		regtmp = regtmp->next;
	}
	return(-1);
}

/**
 * Construct and return a fifo name, something like
 *    /tmp/astaudioout3
 * Return NULL on error.
 */
static char *build_fifo_name(lchan_dir_t direction, int index)
{
	ast_pthread_mutex_lock(&name_lock);
	memset((char *)name, 0, MAX_FIFO_NAME);
	snprintf(name, MAX_FIFO_NAME, "%s%s%d", AST_OH323_FIFO_PREFIX, 
			(direction == RECORDER ? "in" : "out"), index);
	ast_pthread_mutex_unlock(&name_lock);
	return name;
}

/********************************************************************************/
/* Wrapper library's callbacks **************************************************/

/**
 * Call-back function for establishing communication
 * between Asterisk PBX and OpenH323 library.
 * Returns NULL on error, the fifo name on success.
 */
char *setup_h323_connection(call_details_t cd, lchan_dir_t dir, int bs, 
				int codec, int *fd)
{
	int		empty=-1, i;

	if (option_debug) {
		ast_log(LOG_DEBUG, "In setup_h323_connection...\n");
		switch (dir) {
			case RECORDER:
				ast_log(LOG_DEBUG, "RECORDER.\n");
				break;
			case PLAYER:
				ast_log(LOG_DEBUG, "PLAYER.\n");
				break;
			default:
				ast_log(LOG_DEBUG, "W T F??? dir=%d\n", dir);
		}
	}

	ast_pthread_mutex_lock(&oh323_tab_lock);
	*fd = -1;

	/* Search for an already allocated private structure with the */
	/* specified call reference number.*/
	for (i=0; i<config_options.totalNum; i++)
		if (oh323_tab[i] != NULL) {
			if (oh323_tab[i]->cd.call_reference == cd.call_reference)
				if (strcmp(oh323_tab[i]->cd.call_token, cd.call_token) == 0)
					break;
		} else {
			/* Store the first empty record. */
			if (empty == -1)
				empty = i;
		}

	/* No free entry */
	if (((i<0) || (i>=config_options.totalNum)) && (empty == -1)) {
		ast_log(LOG_ERROR, "Unable to accept more calls.\n");
		ast_pthread_mutex_unlock(&oh323_tab_lock);
		return NULL;
	}

	/* Call reference match, the struct is already there. */
	/* Check the direction requested. */
	/* Just return the name of the fifo. */
	if ((i>=0) && (i<config_options.totalNum)) {
		if (oh323_tab[i]->direction == dir) {
			ast_log(LOG_ERROR, "Request to open an existing channel %d with the same direction %d.\n", i, dir);
			ast_pthread_mutex_unlock(&oh323_tab_lock);
			return NULL;
		}
		if (oh323_tab[i]->direction == NONE) {
			if (option_debug)
				ast_log(LOG_DEBUG, "Request to open an existing channel %d with no established direction.\n", i);
			if (copy_call_details(&cd, &(oh323_tab[i]->cd)) != 0){
				ast_log(LOG_ERROR, "Failed to copy call details.\n");
				ast_pthread_mutex_unlock(&oh323_tab_lock);
				return NULL;
			}
			oh323_tab[i]->direction = dir;
			if (dir == PLAYER) {
				oh323_tab[i]->rx_buf_size = bs;
				*fd = oh323_tab[i]->player_fd[1];
			} else if (dir == RECORDER) {
				oh323_tab[i]->tx_buf_size = bs;
				*fd = oh323_tab[i]->recorder_fd[1];
			}
			ast_pthread_mutex_unlock(&oh323_tab_lock);
			return build_fifo_name(dir, i);
		}

		if (option_debug)
			ast_log(LOG_DEBUG, "Call reference match found for %d in %d.\n", cd.call_reference, i);
		oh323_tab[i]->direction = BOTH;
		if (dir == PLAYER) {
			oh323_tab[i]->rx_buf_size = bs;
			*fd = oh323_tab[i]->player_fd[1];
		} else if (dir == RECORDER) {
			oh323_tab[i]->tx_buf_size = bs;
			*fd = oh323_tab[i]->recorder_fd[1];
		}

		/* Set the capability of the channel */
		oh323_tab[i]->capability = oh323_codec2format(codec);
		if (oh323_tab[i]->capability == 0) {
			ast_log(LOG_DEBUG, "Unsupported codec %d!\n", codec);
			ast_pthread_mutex_unlock(&oh323_tab_lock);
			return NULL;
		}
		if (oh323_tab[i]->owner != NULL) {
			oh323_tab[i]->owner->readformat = oh323_tab[i]->capability;
			oh323_tab[i]->owner->writeformat = oh323_tab[i]->capability;
			oh323_tab[i]->owner->nativeformats = oh323_tab[i]->capability;
		}

		/* Set the channel's state */
		if (oh323_tab[i]->from_remote == 1) {
			if (option_debug)
				ast_log(LOG_DEBUG, "Call established (remote).\n");
			oh323_tab[i]->i_state = OH323_STATE_ESTABLISHED;

			if (oh323_tab[i]->owner != NULL)
				ast_setstate(oh323_tab[i]->owner, AST_STATE_UP);
			
			/* Update statistics */
			ast_pthread_mutex_lock(&oh323_stats_lock);
			++oh323_stats.incall;
			ast_pthread_mutex_unlock(&oh323_stats_lock);
		} else {
			if (option_debug)
				ast_log(LOG_DEBUG, "Call established (local).\n");
			oh323_tab[i]->i_state = OH323_STATE_ESTABLISHED;

			/* Update statistics */
			ast_pthread_mutex_lock(&oh323_stats_lock);
			++oh323_stats.outcall;
			ast_pthread_mutex_unlock(&oh323_stats_lock);
		}
		ast_pthread_mutex_unlock(&oh323_tab_lock);
		return build_fifo_name(dir, i);
	}

	ast_log(LOG_ERROR, "XXX NO SUCH CALL REFERENCE!!! XXX\n");
	ast_pthread_mutex_unlock(&oh323_tab_lock);
	return NULL;
}


/**
 * Call-back function to cleanup communication
 * previously established by "setup_h323_connection(...)"
 * Always returns NULL.
 */
char *cleanup_h323_connection(call_details_t cd)
{	
	int i;

	if (option_debug)
		ast_log(LOG_DEBUG, "In cleanup_h323_connection...\n");

	ast_pthread_mutex_lock(&oh323_tab_lock);

	/* Search the already allocated private structures for a */
	/* matching call reference number.*/
	for (i=0; i<config_options.totalNum; i++)
		if (oh323_tab[i] != NULL)
			if (oh323_tab[i]->cd.call_reference == cd.call_reference)
				if (strcmp(oh323_tab[i]->cd.call_token, cd.call_token) == 0)
					break;

	/* No match found */
	if ((i<0) || (i>=config_options.totalNum)) {
		if (option_debug)
			ast_log(LOG_WARNING, "Call reference not found.\n");
		ast_pthread_mutex_unlock(&oh323_tab_lock);
		return NULL;
	}

	/* Call reference match. */
	if (option_debug) {
		ast_log(LOG_DEBUG, "Call reference match found for %d in %d (%d bytes).\n", 
						cd.call_reference, i, sizeof(*(oh323_tab[i])));
	}
	if (oh323_tab[i]->i_state == OH323_STATE_INIT) {
		if (option_debug)
			ast_log(LOG_DEBUG, "Call with reference %d cleared in INIT state.\n",
						cd.call_reference);

		oh323_tab[i]->i_state = OH323_STATE_CLEARED;
		oh323_tab[i]->alreadygone = 1;
		ast_pthread_mutex_unlock(&oh323_tab_lock);
		return NULL;
	}
	if (oh323_tab[i]->owner != NULL) {
		char shutdown_buf[100] = {0};

		if (option_debug)
			ast_log(LOG_DEBUG, "Forcing H.323 channel to hangup.\n");

		oh323_tab[i]->alreadygone = 1;
		ast_setstate(oh323_tab[i]->owner, AST_STATE_DOWN);
		/*oh323_tab[i]->owner->softhangup = 1;*/
		/* XXX Force OpenH323 channel to hangup */
		/*ast_softhangup(oh323_tab[i]->owner, 1);*/
		ast_queue_hangup(oh323_tab[i]->owner, 0);
		write(oh323_tab[i]->player_fd[1], shutdown_buf, 100);
	} else {
		ast_log(LOG_WARNING, "No owner for H.323 channel %d.\n", i);
	}

	oh323_tab[i]->alreadygone = 1;
	ast_pthread_mutex_unlock(&oh323_tab_lock);

	if (option_debug)
		ast_log(LOG_DEBUG, "Leaving cleanup_h323_connection...\n");

	return NULL;
}

/**
 * Call-back function to inform "oh323_call()" that the
 * remote endpoint was found and is ringing.
 * Always returns 0.
 */
int alerted_h323_connection(call_details_t cd)
{	
	int i;

	if (option_debug)
		ast_log(LOG_DEBUG, "In alerted_h323_connection...\n");

	ast_pthread_mutex_lock(&oh323_tab_lock);

	/* Search the already allocated private structures for a */
	/* matching call reference number.*/
	for (i=0; i<config_options.totalNum; i++)
		if (oh323_tab[i] != NULL)
			if (oh323_tab[i]->cd.call_reference == cd.call_reference)
				if (strcmp(oh323_tab[i]->cd.call_token, cd.call_token) == 0)
					break;

	/* No match found */
	if ((i<0) || (i>=config_options.totalNum)) {
		if (option_debug)
			ast_log(LOG_WARNING, "Call reference not found.\n");
		ast_pthread_mutex_unlock(&oh323_tab_lock);
		return(0);
	}

	/* Call reference match. */
	if (option_debug) {
		ast_log(LOG_DEBUG, "Call reference match found for %d in %d (%d bytes).\n", 
						cd.call_reference, i, sizeof(*(oh323_tab[i])));
	}
	if (oh323_tab[i]->i_state == OH323_STATE_INIT) {
		if (option_debug)
			ast_log(LOG_DEBUG, "Call with reference %d alerted in INIT state.\n",
						cd.call_reference);
		oh323_tab[i]->i_state = OH323_STATE_RING;
	} else {
		ast_log(LOG_WARNING, "Call with reference %d in unexpected state (%d).\n",
						cd.call_reference, oh323_tab[i]->i_state);
	}

	if (option_debug)
		ast_log(LOG_DEBUG, "Leaving alerted_h323_connection...\n");

	ast_pthread_mutex_unlock(&oh323_tab_lock);
	return(0);
}

/**
 * Call-back function for initializing a incoming H.323 call
 * inside ASTERISK.
 * Returns 0 on success, -1 on error.
 */
int init_h323_connection(call_details_t cd)
{
	int					i, count;
	struct ast_channel	*c;

	if (option_debug) {
		ast_log(LOG_DEBUG, "In init_h323_connection...\n");
	}

	ast_pthread_mutex_lock(&oh323_tab_lock);

	/* Check the number of active inbound H.323 calls */
	count = 0;
	for (i=0; i<config_options.totalNum; i++)
		if (oh323_tab[i] != NULL)
			if (oh323_tab[i]->from_remote == 1)
				++count;
	if (count >= config_options.inboundMax) {
		ast_log(LOG_WARNING, "Dropping inbound H.323 call due to call-limit violation.\n");

		/* Update statistics */
		ast_pthread_mutex_lock(&oh323_stats_lock);
		++oh323_stats.block_incall;
		ast_pthread_mutex_unlock(&oh323_stats_lock);

		ast_pthread_mutex_unlock(&oh323_tab_lock);
		return(-1);
	}

	/* Search for a free private structure entry */
	for (i=0; i<config_options.totalNum; i++)
		if (oh323_tab[i] == NULL)
			break;

	/* No free entry */
	if ((i<0) || (i>=config_options.totalNum)) {
		ast_log(LOG_ERROR, "Unable to accept more calls.\n");
		ast_pthread_mutex_unlock(&oh323_tab_lock);
		return(-1);
	}

	/* We found a free entry */
	oh323_tab[i] = new_oh323(i);
	if (oh323_tab[i]) {
		if (option_debug) {
			ast_log(LOG_DEBUG, "Created new call structure %d (%d bytes).\n", 
							i, sizeof(*(oh323_tab[i])));
		}
		/* The call was initiated from the remote endpoint */
		oh323_tab[i]->from_remote = 1;

		/* Copy the call details */
		if (copy_call_details(&cd, &(oh323_tab[i]->cd)) != 0){
			ast_log(LOG_ERROR, "Failed to copy call details.\n");
			ast_pthread_mutex_unlock(&oh323_tab_lock);
			return(-1);
		}

		/* Attach the channel to ASTERISK PBX */
		if ((c = ast_oh323_new(oh323_tab[i], AST_STATE_RING))) {
			if (option_debug)
				ast_log(LOG_DEBUG, "Channel created and attached.\n");
			oh323_tab[i]->i_state = OH323_STATE_RINGING;

			/* Update statistics */
			ast_pthread_mutex_lock(&oh323_stats_lock);
			++oh323_stats.setup_recv;
			ast_pthread_mutex_unlock(&oh323_stats_lock);

			ast_pthread_mutex_unlock(&oh323_tab_lock);
			return(0);
		} else {
			oh323_destroy(i);
			free(oh323_tab[i]);
			oh323_tab[i] = NULL;
			ast_pthread_mutex_unlock(&oh323_tab_lock);
			return(-1);
		}
	}
	ast_log(LOG_WARNING, "Failed to create new H.323 private structure %d.\n", i);
	ast_pthread_mutex_unlock(&oh323_tab_lock);
	return(-1);
}

/********************************************************************************/
/* Configuration file parsing ***************************************************/

static int init_registry_info(struct ast_config *cfg, struct oh323_reginfo **reg,
							  char *default_context)
{
	struct ast_variable	*v = NULL;
	struct oh323_reginfo *regtmp, *regprev;
	int nocontext;

	if ((cfg == NULL)||(reg == NULL)||(default_context == NULL))
		return(-1);
	nocontext = 1;
	*reg = NULL;
	regtmp = NULL;
	regprev = NULL;
	v = ast_variable_browse(cfg, "register");
	while (v) {
		if (!strcasecmp(v->name, "context")) {
			if (regtmp != NULL) {
				if (regtmp->alias_num > 0) {
					regtmp->alias = (char **)malloc(regtmp->alias_num * sizeof(char*));
					if (regtmp->alias == NULL) {
						ast_log(LOG_ERROR, "Memory allocation failed.\n");
						return(-1);
					}
					memset(regtmp->alias, 0, regtmp->alias_num * sizeof(char*));
				}
				if (regtmp->prefix_num > 0) {
					regtmp->prefix = (char **)malloc(regtmp->prefix_num * sizeof(char*));
					if (regtmp->prefix == NULL) {
						ast_log(LOG_ERROR, "Memory allocation failed.\n");
						return(-1);
					}
					memset(regtmp->prefix, 0, regtmp->prefix_num * sizeof(char*));
				}
				regprev = regtmp;
			}
			/* Allocate a new struct */
			regtmp = (struct oh323_reginfo *)malloc(sizeof(struct oh323_reginfo));
			if (regtmp == NULL) {
				ast_log(LOG_ERROR, "Memory allocation failed.\n");
				return(-1);
			}
			memset(regtmp, 0, sizeof(struct oh323_reginfo));
			if (*reg == NULL)	/* Save the head of the list */
				*reg = regtmp;
			else				/* Connect the new struct to the end of the list */
				regprev->next = regtmp;
			strncpy(regtmp->context, v->value, AST_MAX_EXTENSION-1);
			nocontext = 0;
		} else if (!strcasecmp(v->name, "alias")) {
			if ((nocontext)&&(regtmp == NULL)) {
				/* Allocate a new struct */
				regtmp = (struct oh323_reginfo *)malloc(sizeof(struct oh323_reginfo));
				if (regtmp == NULL) {
					ast_log(LOG_ERROR, "Memory allocation failed.\n");
					return(-1);
				}
				memset(regtmp, 0, sizeof(struct oh323_reginfo));
				*reg = regtmp; /* Save the head of the list */
				strncpy(regtmp->context, default_context, AST_MAX_EXTENSION-1);
			}
			++regtmp->alias_num;
		} else if (!strcasecmp(v->name, "gwprefix")) {
			if ((nocontext)&&(regtmp == NULL)) {
				/* Allocate a new struct */
				regtmp = (struct oh323_reginfo *)malloc(sizeof(struct oh323_reginfo));
				if (regtmp == NULL) {
					ast_log(LOG_ERROR, "Memory allocation failed.\n");
					return(-1);
				}
				memset(regtmp, 0, sizeof(struct oh323_reginfo));
				*reg = regtmp; /* Save the head of the list */
				strncpy(regtmp->context, default_context, AST_MAX_EXTENSION-1);
			}
			++regtmp->prefix_num;
		} else
			ast_log(LOG_NOTICE, "Ignoring unknown H.323 [register] keyword '%s', line %d.\n", 
								v->name, v->lineno);
		v = v->next;
	}
	/* Update the last struct of the list */
	if (regtmp != NULL) {
		if (regtmp->alias_num > 0) {
			regtmp->alias = (char **)malloc(regtmp->alias_num * sizeof(char*));
			if (regtmp->alias == NULL) {
				ast_log(LOG_ERROR, "Memory allocation failed.\n");
				return(-1);
			}
			memset(regtmp->alias, 0, regtmp->alias_num * sizeof(char*));
		}
		if (regtmp->prefix_num > 0) {
			regtmp->prefix = (char **)malloc(regtmp->prefix_num * sizeof(char*));
			if (regtmp->prefix == NULL) {
				ast_log(LOG_ERROR, "Memory allocation failed.\n");
				return(-1);
			}
			memset(regtmp->prefix, 0, regtmp->prefix_num * sizeof(char*));
		}
		regtmp->next = NULL;
	}

	return(0);
}

static int reload_config()
{
	struct ast_config *cfg = NULL;
	struct ast_variable	*v = NULL;
	struct oh323_reginfo *regtmp;
	struct oh323_codecinfo *codectmp, *codecprev;
	int format, alias_index, prefix_index;
	int codec_found, frames_found;
	int vcodec, flags;
	/* int i; */

	/* Initialize stuff */
	memset((char *)&config_options, 0, sizeof(config_options));
	config_options.listenPort = 1720;
	config_options.connectPort = 1720;
	config_options.fastStart = 1;
	config_options.h245Tunnelling = 1;
	config_options.h245inSetup = 0;
	config_options.inBandDTMF = 0;
	config_options.silenceSuppression = 0;
	config_options.jitterMin = 50;
	config_options.jitterMax = 200;
	config_options.ipTos = 0;
	config_options.outboundMax = 0;
	config_options.inboundMax = 0;
	config_options.totalNum = 0;
	config_options.libTraceLevel = 0;
	strncpy(config_options.libTraceFile, "stdout", 
					sizeof(config_options.libTraceFile)-1);
	config_options.amaFlags = 0;
	config_options.gatekeeperMode = GKMODE_DISABLE;
	config_options.userInputMode = UIMODE_TONE;
	strncpy(config_options.context, "default", AST_MAX_EXTENSION-1);
	oh323_capability = 0;

	/* We must have a config file, otherwise stop */
	cfg = ast_load(config);
	if (!cfg) {
		ast_log(LOG_ERROR, "Unable to load configuration file (%s).\n", config);
		return(-1);
	}

	/* There must be a [general] category, otherwise stop */
	v = ast_variable_browse(cfg, "general");
	if (v == NULL) {
		ast_log(LOG_ERROR, "Category [general] not present in configuration file (%s).\n", config);
		return(-1);
	}
	while (v) {
		if (!strcasecmp(v->name, "listenAddress")) {
			strncpy(config_options.listenAddress, v->value, 
							sizeof(config_options.listenAddress)-1);
		} else if (!strcasecmp(v->name, "listenPort")) {
			config_options.listenPort = (int) strtol(v->value, NULL, 10);
		} else if (!strcasecmp(v->name, "connectPort")) {
			config_options.connectPort = (int) strtol(v->value, NULL, 10);
		} else if (!strcasecmp(v->name, "fastStart")) {
			config_options.fastStart = ast_true(v->value);
		} else if (!strcasecmp(v->name, "h245Tunnelling")) {
			config_options.h245Tunnelling = ast_true(v->value);
		} else if (!strcasecmp(v->name, "h245inSetup")) {
			config_options.h245inSetup = ast_true(v->value);
		} else if (!strcasecmp(v->name, "inBandDTMF")) {
			config_options.inBandDTMF = ast_true(v->value);
		} else if (!strcasecmp(v->name, "silenceSuppression")) {
			config_options.silenceSuppression = ast_true(v->value);
		} else if (!strcasecmp(v->name, "jitterMin")) {
			config_options.jitterMin = (int) strtol(v->value, NULL, 10);
			if (config_options.jitterMin < 20) {
				ast_log(LOG_WARNING, "Invalid min jitter value (<20ms)! Setting it to 20ms.\n");
				config_options.jitterMin = 20;
			} else if (config_options.jitterMin > 10000) {
				ast_log(LOG_WARNING, "Invalid min jitter value (>10s)! Setting it to 10s.\n");
				config_options.jitterMin = 10000;
			}
		} else if (!strcasecmp(v->name, "jitterMax")) {
			config_options.jitterMax = (int) strtol(v->value, NULL, 10);
			if (config_options.jitterMax < 20) {
				ast_log(LOG_WARNING, "Invalid max jitter value (<20ms)! Setting it to 20ms.\n");
				config_options.jitterMax = 20;
			} else if (config_options.jitterMax > 10000) {
				ast_log(LOG_WARNING, "Invalid max jitter value (>10s)! Setting it to 10s.\n");
				config_options.jitterMax = 10000;
			}
		} else if (!strcasecmp(v->name, "ipTos")) {
			if (sscanf(v->value, "%i", &format) == 1)
				config_options.ipTos = format & 0xff;
			else if (!strcasecmp(v->value, "lowdelay"))
				config_options.ipTos = IPTOS_LOWDELAY;
			else if (!strcasecmp(v->value, "throughput"))
				config_options.ipTos = IPTOS_THROUGHPUT;
			else if (!strcasecmp(v->value, "reliability"))
				config_options.ipTos = IPTOS_RELIABILITY;
			else if (!strcasecmp(v->value, "mincost"))
				config_options.ipTos = IPTOS_MINCOST;
			else if (!strcasecmp(v->value, "none"))
				config_options.ipTos = 0;
			else
				ast_log(LOG_WARNING, "Invalid ToS value (%s) at line %d, should be 'lowdelay', 'throughput', 'reliability', 'mincost', or 'none'.\n",
								v->value, v->lineno);
		} else if (!strcasecmp(v->name, "outboundMax")) {
			config_options.outboundMax = (int) strtol(v->value, NULL, 10);
			if (config_options.outboundMax < 0) {
				ast_log(LOG_WARNING, "Invalid max number of outbound H.323 connections (<0)! Setting it to 0.\n");
				config_options.outboundMax = 0;
			}
		} else if (!strcasecmp(v->name, "inboundMax")) {
			config_options.inboundMax = (int) strtol(v->value, NULL, 10);
			if (config_options.inboundMax < 0) {
				ast_log(LOG_WARNING, "Invalid max number of inbound H.323 connections (<0)! Setting it to 0.\n");
				config_options.inboundMax = 0;
			}
		} else if (!strcasecmp(v->name, "libTraceLevel")) {
			config_options.libTraceLevel = (int) strtol(v->value, NULL, 10);
			if (config_options.libTraceLevel < 0) {
				ast_log(LOG_WARNING, "Invalid library trace level! Disabling it.\n");
				config_options.libTraceLevel = 0;
			}
		} else if (!strcasecmp(v->name, "libTraceFile")) {
			if ((v->value[0] == '\0') || (!strcasecmp(v->value, "stdout")))
				memset(config_options.libTraceFile, 0, sizeof(config_options.libTraceFile));
			else
				strncpy(config_options.libTraceFile, v->value, 
								sizeof(config_options.libTraceFile)-1);
		} else if (!strcasecmp(v->name, "accountCode")) {
			strncpy(config_options.accountCode, v->value, 
								sizeof(config_options.accountCode)-1);
		} else if (!strcasecmp(v->name, "amaFlags")) {
			flags = ast_cdr_amaflags2int(v->value);
			if (flags < 0) {
				ast_log(LOG_WARNING, "Invalid AMA Flags: %s at line %d\n", v->value, v->lineno);
			} else {
				config_options.amaFlags = flags;
			}
		} else if (!strcasecmp(v->name, "gatekeeper")) {
			if (!strcasecmp(v->value, "DISABLE")) {
				config_options.gatekeeperMode = GKMODE_DISABLE;
			} else if (!strcasecmp(v->value, "DISCOVER")) {
				config_options.gatekeeperMode = GKMODE_DISCOVER;
			} else if (!strncasecmp(v->value, "GKID:", 5)){
				config_options.gatekeeperMode = GKMODE_ID;
				strncpy(config_options.gatekeeperName, (char*)(v->value+5), 
								sizeof(config_options.gatekeeperName)-1);
			} else {
				config_options.gatekeeperMode = GKMODE_NAME;
				strncpy(config_options.gatekeeperName, v->value, 
								sizeof(config_options.gatekeeperName)-1);
			}
		} else if (!strcasecmp(v->name, "gatekeeperPassword")) {
			strncpy(config_options.gatekeeperPass, v->value, 
							sizeof(config_options.gatekeeperPass)-1);
		} else if (!strcasecmp(v->name, "userInputMode")) {
			if (!strcasecmp(v->value, "q931")) {
				config_options.userInputMode = UIMODE_Q931;
			} else if (!strcasecmp(v->value, "string")) {
				config_options.userInputMode = UIMODE_STRING;
			} else if (!strcasecmp(v->value, "tone")) {
				config_options.userInputMode = UIMODE_TONE;
			} else if (!strcasecmp(v->value, "rfc2833")) {
				config_options.userInputMode = UIMODE_RFC2833;
			} else
				ast_log(LOG_WARNING, "%s: Unknown user-input mode.\n", v->value);
		} else if (!strcasecmp(v->name, "context")) {
			strncpy(config_options.context, v->value, sizeof(config_options.context)-1);
		} else
			ast_log(LOG_NOTICE, "Ignoring unknown H.323 [general] keyword '%s', line %d.\n", 
								v->name, v->lineno);
		v = v->next;
	}

	if (config_options.jitterMin > config_options.jitterMax) {
		ast_log(LOG_WARNING, "Invalid jitter buffer values (min=%d, max=%d). Resetting to defaults.\n",
						config_options.jitterMin, config_options.jitterMax);
		config_options.jitterMin = 100;
		config_options.jitterMax = 500;
	}
	/* XXX We should restrict this number up to an upper limit */
	config_options.totalNum = config_options.inboundMax+config_options.outboundMax;

	/* Check for a [register] category */
	if (init_registry_info(cfg, &(config_options.regInfo), config_options.context) != 0)
		return(-1);
	if (config_options.regInfo) {
		regtmp = config_options.regInfo;
		alias_index = 0;
		prefix_index = 0;
		v = ast_variable_browse(cfg, "register");
		while (v) {
			if ((alias_index == regtmp->alias_num)&&(prefix_index == regtmp->prefix_num)) {
				regtmp = regtmp->next;
				alias_index = 0;
				prefix_index = 0;
			}
			if (!strcasecmp(v->name, "alias")) {
				regtmp->alias[alias_index] = (char *)malloc(strlen(v->value)+1);
				if (regtmp->alias[alias_index] == NULL) {
					ast_log(LOG_ERROR, "Memory allocation failed.\n");
					return(-1);
				}
				memset(regtmp->alias[alias_index], 0, strlen(v->value)+1);
				strcpy(regtmp->alias[alias_index], v->value);
				++alias_index;
			} else if (!strcasecmp(v->name, "gwprefix")) {
				regtmp->prefix[prefix_index] = (char *)malloc(strlen(v->value)+1);
				if (regtmp->prefix[prefix_index] == NULL) {
					ast_log(LOG_ERROR, "Memory allocation failed.\n");
					return(-1);
				}
				memset(regtmp->prefix[prefix_index], 0, strlen(v->value)+1);
				strcpy(regtmp->prefix[prefix_index], v->value);
				++prefix_index;
			}
			v = v->next;
		}
	} else {
		ast_log(LOG_DEBUG, "H.323 [register] category contains no entries.\n");
	}

	/* There must be a [codecs] category, otherwise stop */
	codec_found = 0;
	frames_found = 0;
	codectmp = NULL;
	config_options.codecInfo = NULL;
	v = ast_variable_browse(cfg, "codecs");
	if (v == NULL) {
		ast_log(LOG_WARNING, "Category [codecs] not present in configuration file (%s).\n", config);
		ast_destroy(cfg);
		return(0);
	}
	while (v) {
		if (!strcasecmp(v->name, "codec")) {
			vcodec = oh323_str2codec(v->value);
			if (vcodec == 0) {
				ast_log(LOG_WARNING, "Unknown codec '%s'.\n", v->value);
				codec_found = 0;
			} else {
				format = oh323_codec2format(vcodec);
				if (format == 0) {
					ast_log(LOG_WARNING, "Unsupported codec '%s'.\n", v->value);
					codec_found = 0;
				} else {
					oh323_capability |= format;

					/* Allocate a new struct */
					if (codectmp != NULL)
						codecprev = codectmp;
					codectmp = (struct oh323_codecinfo*)malloc(sizeof(struct oh323_codecinfo));
					if (codectmp == NULL) {
						ast_log(LOG_ERROR, "Memory allocation failed.\n");
						return(-1);
					}
					memset(codectmp, 0, sizeof(struct oh323_codecinfo));
					codectmp->codec = vcodec;
					codectmp->format = format;
					codectmp->frames = 1;
					if (config_options.codecInfo == NULL)	/* Save the head of the list */
						config_options.codecInfo = codectmp;
					else	 /* Connect the new struct to the end of the list */
						codecprev->next = codectmp;
					codec_found = 1;
					frames_found = 0;
				}
			}
		} else if (!strcasecmp(v->name, "frames")) {
			if ((!codec_found)||(frames_found))
				ast_log(LOG_NOTICE, "Ignoring unexpected 'frames' options at line %d.\n", v->lineno);
			else {
				codectmp->frames = (int) strtol(v->value, NULL, 10);
				frames_found = 1;
				codec_found = 0;
			}
		} else
			ast_log(LOG_NOTICE, "Ignoring unknown H.323 [codecs] keyword '%s'.\n", v->name);
		v = v->next;
	}
	ast_destroy(cfg);

	/******** DEBUG START *****
	regtmp = config_options.regInfo;
	while (regtmp) {
		printf("Context: %s\n", regtmp->context);
		for (i=0; i<regtmp->alias_num; i++)
			printf("   Alias: %s\n", regtmp->alias[i]);
		for (i=0; i<regtmp->prefix_num; i++)
			printf("   Prefix: %s\n", regtmp->prefix[i]);
		regtmp = regtmp->next;
	}
	codectmp = config_options.codecInfo;
	while (codectmp) {
		printf("Codec: %d\n", codectmp->codec);
		printf("   Format: %d\n", codectmp->format);
		printf("   Frames: %d\n", codectmp->frames);
		codectmp = codectmp->next;
	}
	******** DEBUG END *****/

	return(0);
}

static void destroy_config()
{
	struct oh323_reginfo *regtmp, *regtmp1;
	struct oh323_codecinfo *codectmp, *codectmp1;
	int i;

	regtmp = config_options.regInfo;
	while (regtmp) {
		for (i=0; i<regtmp->alias_num; i++)
			if (regtmp->alias[i])
				free(regtmp->alias[i]);
		for (i=0; i<regtmp->prefix_num; i++)
			if (regtmp->prefix[i])
				free(regtmp->prefix[i]);
		regtmp1 = regtmp->next;
		free(regtmp);
		regtmp = regtmp1;
	}
	codectmp = config_options.codecInfo;
	while (codectmp) {
		codectmp1 = codectmp->next;
		free(codectmp);
		codectmp = codectmp1;
	}
	memset((char *)&config_options, 0, sizeof(config_options));
}

/********************************************************************************/
/* Required ASTERISK's stuff ****************************************************/

int load_module()
{
	char **gateway_prefix_tab, **alias_tab;
	struct oh323_reginfo *regtmp;
	struct oh323_codecinfo *codectmp;
	int i, j, k, prefix_num, alias_num;
	int res;

	/* Register our CLI extensions */
	if (option_debug)
		ast_log(LOG_DEBUG, "Registering CLI extensions.\n");
	ast_cli_register(&cli_show_conf);
	ast_cli_register(&cli_show_stats);
	ast_cli_register(&cli_show_info);
	ast_cli_register(&cli_debug_toggle);

	/* Load our config file */
	if (option_debug)
		ast_log(LOG_DEBUG, "Loading config file.\n");
	if (reload_config() < 0)
		return(-1);

	if (config_options.totalNum == 0) {
		ast_log(LOG_NOTICE, "Total number of allowed H.323 calls is 0! Disabling H.323 channel driver.\n");
		return(0);
	}
	if (config_options.codecInfo == NULL) {
		ast_log(LOG_NOTICE, "No codecs configured! Disabling H.323 channel driver.\n");
		return(0);
	}

	/* Init table with private structures of H.323 calls. */
	if (option_debug)
		ast_log(LOG_DEBUG, "Allocating H.323 channel space.\n");
	oh323_tab = (struct chan_oh323_pvt **)malloc(config_options.totalNum * 
												sizeof(struct chan_oh323_pvt *));
	if (oh323_tab == NULL) {
		ast_log(LOG_ERROR, "Memory allocation failed.\n");
		return(-1);
	}
	memset(oh323_tab, 0, config_options.totalNum * sizeof(struct chan_oh323_pvt *));

	/* Init statistics */
	if (option_debug)
		ast_log(LOG_DEBUG, "Initializing statistics.\n");
	ast_pthread_mutex_lock(&oh323_stats_lock);
	oh323_stats.incall = 0;
	oh323_stats.outcall = 0;
	oh323_stats.setup_sent = 0;
	oh323_stats.setup_recv = 0;
	oh323_stats.block_incall = 0;
	oh323_stats.block_outcall = 0;
	ast_pthread_mutex_unlock(&oh323_stats_lock);

	/* Get aliases and prefixes */
	if (option_debug)
		ast_log(LOG_DEBUG, "Initializing aliases/prefixes.\n");
	gateway_prefix_tab = NULL;
	alias_tab = NULL;
	regtmp = config_options.regInfo;
	prefix_num = 0;
	alias_num = 0;
	while (regtmp) {
		prefix_num += regtmp->prefix_num;
		alias_num += regtmp->alias_num;
		regtmp = regtmp->next;
	}
	if (prefix_num > 0) {
		gateway_prefix_tab = (char **)malloc(prefix_num * sizeof(char *));
		if (gateway_prefix_tab == NULL) {
			ast_log(LOG_ERROR, "Memory allocation failed.\n");
			return(-1);
		}
		memset(gateway_prefix_tab, 0, prefix_num * sizeof(char *));
	} else {
		ast_log(LOG_DEBUG, "Zero prefix(es) for ASTERISK.\n");
	}
	if (alias_num > 0) {
		alias_tab = (char **)malloc(alias_num * sizeof(char *));
		if (alias_tab == NULL) {
			ast_log(LOG_ERROR, "Memory allocation failed.\n");
			return(-1);
		}
		memset(alias_tab, 0, alias_num * sizeof(char *));
	} else {
		ast_log(LOG_DEBUG, "Zero alias(es) for ASTERISK.\n");
	}

	regtmp = config_options.regInfo;
	j = 0;
	k = 0;
	while (regtmp) {
		for (i=0; i<regtmp->prefix_num; i++) {
			gateway_prefix_tab[j] = regtmp->prefix[i];
			++j;
		}
		for (i=0; i<regtmp->alias_num; i++) {
			alias_tab[k] = regtmp->alias[i];
			++k;
		}
		regtmp = regtmp->next;
	}

	/* Create our H.323 endpoint */
	/*ast_log(LOG_DEBUG, "Initializing endpoint\n");*/
	if (option_debug)
		ast_log(LOG_DEBUG, "Initializing endpoint.\n");
	h323_end_point_create(gateway_prefix_tab, prefix_num, 
					config_options.libTraceLevel, config_options.libTraceFile);

	/* Configure the endpoint */
	if (option_debug)
		ast_log(LOG_DEBUG, "Configuring endpoint.\n");
	if (h323_set_options(!config_options.fastStart,
						!config_options.h245Tunnelling,
						!config_options.h245inSetup,
						!config_options.silenceSuppression,
						config_options.jitterMin,
						config_options.jitterMax) < 0) {
		ast_log(LOG_ERROR, "H.323 options setup failed.\n");
		unload_module();
		return(-1);
	}

	/* Create the H.323 listener */
	if (option_debug)
		ast_log(LOG_DEBUG, "Starting listener.\n");
	if (h323_start_listener(TCP, config_options.listenAddress,
							config_options.listenPort) != LIS_START_OK) {
		ast_log(LOG_ERROR, "H.323 listener creation failed.\n");
		unload_module();
		return(-1);
	}

	/* Gatekeeper setup */
	if (option_debug)
		ast_log(LOG_DEBUG, "Gatekeeper setup.\n");
	if (h323_set_gk(config_options.gatekeeperMode, config_options.gatekeeperName,
					config_options.gatekeeperPass, alias_tab, alias_num)) {
		ast_log(LOG_ERROR, "H.323 gatekeeper setup failed.\n");
		unload_module();
		return(-1);
	}

	/* Add the codecs */
	/* Note: 
	 * 		AddCapability puts the codec into the list of codecs we can send
	 * 		SetCapability puts the codec into the list of codecs we can send and receive
	 */
	if (option_debug)
		ast_log(LOG_DEBUG, "Capability setup.\n");
	codectmp = config_options.codecInfo;
	while (codectmp) {
		res = h323_set_capability(codectmp->codec, codectmp->frames);
		if (res != CAP_INSERT_OK) {
			ast_log(LOG_ERROR, "Failed to insert capability %d.\n", codectmp->codec);
			unload_module();
			return(-1);
		}
		codectmp = codectmp->next;
	}

	/* Set the user-input mode */
	if (option_debug)
		ast_log(LOG_DEBUG, "User-input mode setup.\n");
	if (h323_set_senduimode(config_options.userInputMode) != CAP_INSERT_OK) {
		ast_log(LOG_ERROR, "Failed to set user-input mode to %d.\n", config_options.userInputMode);
		unload_module();
		return(-1);
	}

	/* Register our callback functions */
	if (option_debug)
		ast_log(LOG_DEBUG, "Callback functions setup.\n");
	if (h323_callback_register(setup_h323_connection, cleanup_h323_connection, 
					alerted_h323_connection, activate_exception, 
					init_h323_connection) < 0) {
		ast_log(LOG_ERROR, "Failed to register callback function(s).\n");
		unload_module();
		return(-1);
	}
	
	/* Register this channel type with Asterisk PBX. */
	if (option_debug)
		ast_log(LOG_DEBUG, "Channel registration.\n");
	if (ast_channel_register(type, tdesc, oh323_capability, oh323_request)) {
		ast_log(LOG_ERROR, "Unable to register channel class %s\n", type);
		unload_module();
		return(-1);
	}

	if (option_verbose > 1)
		ast_verbose(VERBOSE_PREFIX_2 "OpenH323 Channel Ready (v%d.%d.%d)\n", 
						OH323_VERSION_MAJOR, OH323_VERSION_MINOR, OH323_VERSION_BUILD);

	return(0);
}

int unload_module()
{
	int		i;

	if (option_debug)
		ast_log(LOG_DEBUG, "Cleaning up OpenH323 channel driver...\n");

	if (option_debug)
		ast_log(LOG_DEBUG, "Unregistering CLI extensions.\n");
	ast_cli_unregister(&cli_show_conf);
	ast_cli_unregister(&cli_show_stats);
	ast_cli_unregister(&cli_show_info);
	ast_cli_unregister(&cli_debug_toggle);

	/* Unregister channel type */
	if (option_debug)
		ast_log(LOG_DEBUG, "Unregistering channel type.\n");
	ast_channel_unregister(type);

	/* Cleanup stuff */
	if (option_debug)
		ast_log(LOG_DEBUG, "Freeing up resources.\n");
	for (i=0; i<config_options.totalNum; i++)
		if (oh323_tab[i] != NULL) {
			oh323_destroy(i);
			free(oh323_tab[i]);
			oh323_tab[i] = NULL;
		}

	if (option_debug)
		ast_log(LOG_DEBUG, "Removing capabilities.\n");
	if (h323_removeall_capabilities() != CAP_REMOVEALL_OK)
		ast_log(LOG_ERROR, "Unable to remove H323 capabilities.\n");
	if (option_debug)
		ast_log(LOG_DEBUG, "Removing listener.\n");
	if (h323_removeall_listeners() != LIS_REMOVEALL_OK)
		ast_log(LOG_ERROR, "Unable to remove H323 listeners.\n");

	if (option_debug)
		ast_log(LOG_DEBUG, "Destroying endpoint.\n");
	h323_end_point_destroy();
	destroy_config();

	if (option_debug)
		ast_log(LOG_DEBUG, "Done...\n");

	return(0);
}

int usecount()
{
	int res;
	ast_pthread_mutex_lock(&usecnt_lock);
	res = usecnt;
	ast_pthread_mutex_unlock(&usecnt_lock);
	return res;
}

char *description()
{
	return desc;
}

char *key()
{
	return ASTERISK_GPL_KEY;
}

/* End of file *************************************************************/
