/*
 * wrapper.hxx
 *
 * The header file for the OpenH323 Wrapper Library
 *
 * Copyright (c) 2002,2003 InAccess Networks
 * Michalis Manousos <manousos@inaccessnetworks.com>
 * Dimitris Economou <decon@inaccessnetworks.com>
 *
 * This file is part of "H323 support for ASTERISK"
 *
 * "H323 support for ASTERISK" is free software;
 * you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any later version. 
 *
 * "H323 support for ASTERISK" is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details. 
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 * $Id: wrapper.hxx,v 1.20 2003/03/17 17:16:31 manousos Exp $
 *
 */

/************************************************************************/

#ifndef	_WRAPPER_HXX
#define	_WRAPPER_HXX

/************************************************************************/
/* TYPES & VARIABLES ****************************************************/

/* Different types of exceptions inside the OpenH323 Library
   handled by ASTERISK. */
#define	OH323EXC_NULL				0x0
#define	OH323EXC_USER_INPUT_TONE	0x1

/* The boolean type */
typedef enum boolean {
	YES,
	NO    
} boolean_t;

/* The codec capability enumarator */
typedef enum cap_type {
	G711U = 1,	/* G.711 u-Law */
	G711A,		/* G.711 A-Law */
	G7231,		/* G.723.1(6.3k) */
	G72316K3,	/* G.723.1(6.3k) */
	G72315K3,	/* G.723.1(5.3k) */
	G7231A6K3,	/* G.723.1A(6.3k) */
	G7231A5K3,	/* G.723.1A(5.3k) */
	G728,		/* G.728 */
	G729,		/* G.729 */
	G729A,		/* G.729A */
	G729B,		/* G.729B */
	G729AB,		/* G.729AB */
	GSM0610,	/* GSM 0610 */
	MSGSM,		/* Microsoft GSM Audio Capability */
	LPC10,		/* LPC-10 */
	LINEAR16,	/* 16 bit linear PCM */
	CAP_UNDEFINED
} cap_type_t;

typedef enum cap_ret_val {
	CAP_EP_ER,
	CAP_NSUP_ER,
	CAP_EXIST,
	CAP_INSERT_OK,
	CAP_REMOVE_OK,
	CAP_REMOVEALL_OK
} cap_ret_val_t;

/* The user input mode enumarator */
typedef enum ui_mode {
	UIMODE_Q931,		/* Q.931 Keypad Information Element */
	UIMODE_STRING,		/* H.245 string */
	UIMODE_TONE,		/* H.245 tone */
	UIMODE_RFC2833,	/* RFC2833 */
	UIMODE_UNDEFINED
} uimode_t;

typedef enum gk_mode {
	GKMODE_DISABLE = 0,
	GKMODE_DISCOVER,
	GKMODE_NAME,
	GKMODE_ID
} gkmode_t;

/* The listener enumarator 
   This enumerator could hold all the
   possible types of interfaces (i.e. ATM) */
typedef enum lis_type {
	TCP,
	UDP,
	LIS_UNDEFINED
} lis_type_t;

/* These are the return values for
   the h323_start_listener and
   the h323_remove_listener functions. */
typedef enum lis_ret_val {
	LIS_EP_ER,			/* End Point error */
	LIS_NSUP_ER,		/* Listener not supported */
	LIS_EXIST,			/* Listener already exist */
	LIS_FAILOP_ER,		/* Fail to establish listener */
	LIS_START_OK,		/* Listener is started succesfully */
	LIS_REMOVE_OK,		/* Listener is removed succesfully */
	LIS_REMOVEALL_OK	/* All listeners were removed succesfully */
} lis_ret_val_t;

/* This enum shows the direction of a logical channel. */
typedef enum lchan_dir {
	NONE = 0,
	PLAYER,
	RECORDER,
	BOTH
} lchan_dir_t;

/* The return values of the h323_{make/clear/answer}_call functions */
typedef enum call_ret_val {
	CALL_EP_ER,
	CALL_START_OK,
	CALL_START_ER,
	CALL_END_OK,
	CALL_END_ER,
	CALL_ANS_OK,
	CALL_ANS_ER
} call_ret_val_t;

/* The call_option struct is filled from the 
   PBX application and passed through make_call 
   function*/
typedef struct call_options {
/*	boolean_t   noFastStart;
	boolean_t   noH245Tunnelling;
	boolean_t   noSilenceSuppression;
	int			jitter;*/
	char			*cid;
	cap_type_t		cap;
	unsigned int	connectPort;
} call_options_t;

typedef struct call_details {
	const char *call_token;
	unsigned int call_reference;
	const char *call_source_alias;
	const char *call_dest_alias;
	const char *call_source_e164;
	const char *call_dest_e164;
} call_details_t;

/* This is a callback prototype function, called when the openh323 
   OnStartLogicalChannel is invoked. */
typedef char *(*start_logchan_cb)(call_details_t, lchan_dir_t, int, int, int *);

/* This is the callback prototype function which has to be called when
   the openH323 OnConnectionCleared callback is invoked */
typedef char *(*clear_con_cb)(call_details_t);

/* This is the callback prototype function which has to be
   called when the OnAlerting callback is invoked */
typedef int (*alert_con_cb)(call_details_t);

/* This is the callback prototype function called when the openh323
   library has an exception pending. */
typedef int (*h323_exception_cb)(unsigned int, int, char *);

/* This is the callback prototype function called when the
   OnAnswerCall callback is invoked. */
typedef int (*init_con_cb)(call_details_t);

/************************************************************************/
/* FUNCTIONS ************************************************************/

#ifdef cplusplus
extern "C" {
#endif   

	/* See "cap_type" above */
	const int h323_codec_number = 16;

	void h323_end_point_create(char **, int, int, char *);
	void h323_end_point_destroy(void);
	boolean_t end_point_exist(void);
    
	/* Set the callback functions */
	int h323_callback_register(start_logchan_cb, clear_con_cb, alert_con_cb,
									h323_exception_cb, init_con_cb);

	/* H.323 codec and user-input capability related functions */
	cap_ret_val_t	h323_add_capability(cap_type_t, int);
	cap_ret_val_t	h323_set_capability(cap_type_t, int);
	cap_ret_val_t	h323_find_capability(cap_type_t);
	cap_ret_val_t	h323_removeall_capabilities();
	cap_ret_val_t	h323_set_senduimode(uimode_t);

	/* H.323 options setup related functions */
	int				h323_set_options(int, int, int,	int, int, int);

	/* H.323 listener related funcions */
	lis_ret_val_t 	h323_start_listener(lis_type_t lis, char *, int);
	lis_ret_val_t 	h323_remove_listener(lis_type_t lis);
	lis_ret_val_t 	h323_removeall_listeners();
	
	/* H.323 gatekeeper related funcions */
	char			*h323_get_gk(char *);
	int				h323_set_gk(gkmode_t, char *, char *, char **, int);

	/* Send a DTMF tone to remote endpoint */
	void h323_send_tone(const char *call_token, char tone);

	/* H.323 create and destroy sessions */
	call_ret_val_t h323_make_call(char *host, call_details_t *pcd, 
									call_options_t call_options);
	call_ret_val_t h323_clear_call(const char *call_token);
	call_ret_val_t h323_answer_call(const char *call_token);

#ifdef cplusplus
}
#endif

#endif	/* _WRAPPER_HXX */

/* End of file *************************************************************/
